import numpy as np
from numpy import array, nanmin, nanmax
from sympy import Symbol
from pysces import ModelMap, Scanner, ParScanner
from numpy import NaN, abs

from ...utils.model_graph import ModelGraph
from ...utils.misc import silence_print, DotDict, formatter_factory, \
    do_safe_state, find_min, find_max, get_value, stringify, \
    scanner_range_setup

from ...utils.plotting import Data2D


def cctype(obj):
    return 'ccobjects' in str(type(obj))


@silence_print
def get_state(mod, do_state=False):
    if do_state:
        mod.doState()
    ss = [getattr(mod, 'J_' + r) for r in mod.reactions] + \
        [getattr(mod, s + '_ss') for s in mod.species]
    return ss


class CCBase(object):

    """The base object for the control coefficients and control patterns"""

    def __init__(self, mod, name, expression, ltxe):
        super(CCBase, self).__init__()

        self.expression = expression
        self.mod = mod
        self._ltxe = ltxe
        self.name = name
        self._latex_name = '\\Sigma'
        self._analysis_method = 'symca'

        self._str_expression_ = None
        self._value = None
        self._latex_expression = None

    @property
    def latex_expression(self):
        if not self._latex_expression:
            self._latex_expression = self._ltxe.expression_to_latex(
                self.expression
            )
        return self._latex_expression

    @property
    def latex_name(self):
        return self._latex_name

    @property
    def _str_expression(self):
        if not self._str_expression_:
            self._str_expression_ = str(self.expression)
        return self._str_expression_

    @property
    def value(self):
        """The value property. Calls self._calc_value() when self._value
        is None and returns self._value"""
        self._calc_value()
        return self._value

    def _repr_latex_(self):
        return '$%s = %s = %.3f$' % (self.latex_name,
                                     self.latex_expression,
                                     self.value)

    def _calc_value(self):
        """Calculates the value of the expression"""
        keys = self.expression.atoms(Symbol)
        subsdict = {}
        for key in keys:
            str_key = str(key)
            subsdict[str_key] = getattr(self.mod, str_key)
        self._value = get_value(self._str_expression, subsdict)

    def __repr__(self):
        return self.expression.__repr__()

    def __add__(self, other):
        if cctype(other):
            return self.expression.__add__(other.expression)
        else:
            return self.expression.__add__(other)

    def __mul__(self, other):
        if cctype(other):
            return self.expression.__mul__(other.expression)
        else:
            return self.expression.__mul__(other)

    def __div__(self, other):
        if cctype(other):
            return self.expression.__div__(other.expression)
        else:
            return self.expression.__div__(other)

    def __pow__(self, other):
        if cctype(other):
            return self.expression.__pow__(other.expression)
        else:
            return self.expression.__pow__(other)


class CCoef(CCBase):

    """The object the stores control coefficients. Inherits from CCBase"""

    def __init__(self, mod, name, expression, denominator, ltxe):
        super(CCoef, self).__init__(mod, name, expression, ltxe)
        self.numerator = expression
        self.denominator = denominator.expression
        self.expression = self.numerator / denominator.expression
        self.denominator_object = denominator

        self._latex_numerator = None
        self._latex_expression_full = None
        self._latex_expression = None
        self._latex_name = None
        self._abs_value = None

        self.control_patterns = None

        self._set_control_patterns()

    @property
    def abs_value(self):
        self._calc_abs_value()
        return self._abs_value

    @property
    def latex_numerator(self):
        if not self._latex_numerator:
            self._latex_numerator = self._ltxe.expression_to_latex(
                self.numerator
            )
        return self._latex_numerator

    @property
    def latex_expression_full(self):
        if not self._latex_expression_full:
            full_expr = '\\frac{' + self.latex_numerator + '}{' \
                + self.denominator_object.latex_expression + '}'
            self._latex_expression_full = full_expr
        return self._latex_expression_full

    @property
    def latex_expression(self):
        if not self._latex_expression:
            self._latex_expression = '(' + \
                self.latex_numerator + ')' + '/~\\Sigma'
        return self._latex_expression

    @property
    def latex_name(self):
        if not self._latex_name:
            self._latex_name = self._ltxe.expression_to_latex(
                self.name
            )
        return self._latex_name

    def _perscan_legacy(self, parameter, scan_range):

        scan_res = [list()
                    for i in range(len(self.control_patterns.values()) + 1)]
        scan_res[0] = scan_range

        for parvalue in scan_range:
            state_valid = do_safe_state(
                self.mod, parameter, parvalue, type='mca')
            cc_abs_value = 0
            for i, cp in enumerate(self.control_patterns.values()):
                if state_valid:
                    cp_abs = abs(cp.value)
                    scan_res[i + 1].append(cp_abs)
                    cc_abs_value += cp_abs
                else:
                    scan_res[i + 1].append(NaN)

            for i, cp in enumerate(self.control_patterns.values()):
                if state_valid:
                    scan_res[i + 1][-1] = (scan_res[i + 1]
                                           [-1] / cc_abs_value) * 100

        return scan_res

    def _valscan_legacy(self, parameter, scan_range):

        control_pattern_range = range(len(self.control_patterns.values()) + 2)
        scan_res = [list() for i in control_pattern_range]
        scan_res[0] = scan_range

        for parvalue in scan_range:
            state_valid = do_safe_state(self.mod,
                                        parameter,
                                        parvalue,
                                        type='mca')
            cc_value = 0
            for i, cp in enumerate(self.control_patterns.values()):
                if state_valid:
                    cp_value = cp.value
                    scan_res[i + 1].append(cp_value)
                    cc_value += cp_value
                else:
                    scan_res[i + 1].append(NaN)
            if state_valid:
                scan_res[i + 2].append(cc_value)
            else:
                scan_res[i + 2].append(NaN)

        return scan_res

    def _perscan(self,
                 parameter,
                 scan_range,
                 par_scan=False,
                 par_engine='multiproc'):

        val_scan_res = self._valscan(parameter,
                                     scan_range,
                                     par_scan,
                                     par_engine)

        points = len(scan_range)
        parameter = val_scan_res[:, 0].reshape(points, 1)
        cp_abs_vals = np.abs(val_scan_res[:, 1:-1])
        cp_abs_sum = np.sum(cp_abs_vals, 1).reshape(points, 1)
        cp_abs_perc = (cp_abs_vals / cp_abs_sum) * 100
        scan_res = np.hstack([parameter, cp_abs_perc])
        return scan_res

    def _valscan(self,
                 parameter,
                 scan_range,
                 par_scan=False,
                 par_engine='multiproc'):

        needed_symbols = [parameter] + \
            stringify(list(self.expression.atoms(Symbol)))

        # This is experimental
        if par_scan:
            scanner = ParScanner(self.mod, par_engine)
        else:
            scanner = Scanner(self.mod)
            scanner.quietRun = True

        start, end, points, log = scanner_range_setup(scan_range)
        scanner.addScanParameter(parameter,
                                 start=start,
                                 end=end,
                                 points=points,
                                 log=log)
        scanner.addUserOutput(*needed_symbols)
        scanner.Run()
        subs_dict = {}
        for i, symbol in enumerate(scanner.UserOutputList):
            subs_dict[symbol] = scanner.UserOutputResults[:, i]

        control_pattern_names = self.control_patterns.keys()
        denom_expr = str(self.denominator)
        cp_numerators = [self.control_patterns[cp_name].numerator for
                         cp_name in control_pattern_names]
        column_exprs = stringify(cp_numerators)

        parameter = subs_dict[parameter].reshape(points, 1)

        scan_res = []

        denom_val = get_value(denom_expr, subs_dict)
        for expr in column_exprs:
            scan_res.append(get_value(expr, subs_dict) / denom_val)
        scan_res = np.array(scan_res).transpose()
        cc_vals = np.sum(scan_res, 1).reshape(points, 1)
        scan_res = np.hstack([parameter, scan_res, cc_vals])
        return scan_res

    def do_par_scan(self,
                    parameter,
                    scan_range,
                    scan_type='percentage',
                    init_return=True,
                    par_scan=False,
                    par_engine='multiproc',
                    force_legacy=False):

        assert scan_type in ['percentage', 'value']
        init = getattr(self.mod, parameter)

        column_names = [parameter] + \
                       [cp.name for cp in self.control_patterns.values()]

        if scan_type is 'percentage':
            y_label = 'Control pattern percentage contribution'
            try:
                assert not force_legacy, 'Legacy scan requested'
                scan_res = self._perscan(parameter,
                                         scan_range,
                                         par_scan,
                                         par_engine)
                data_array = scan_res
            except Exception as exception:
                print 'The parameter scan yielded the following error:'
                print exception
                print 'Switching over to slower scan method and replacing'
                print 'invalid steady states with NaN values.'
                scan_res = self._perscan_legacy(parameter, scan_range)
                data_array = array(scan_res, dtype=np.float).transpose()

            ylim = [nanmin(data_array[:, 1:]), nanmax(data_array[:, 1:]) * 1.1]
        elif scan_type is 'value':
            column_names = column_names + [self.name]
            y_label = 'Control coefficient/pattern value'
            try:
                assert not force_legacy, 'Legacy scan requested'
                scan_res = self._valscan(parameter,
                                         scan_range,
                                         par_scan,
                                         par_engine)
                data_array = scan_res
            except Exception as exception:
                print 'The parameter scan yielded the following error:'
                print exception
                print 'Switching over to slower scan method and replacing'
                print 'invalid steady states with NaN values.'
                scan_res = self._valscan_legacy(parameter, scan_range)
                data_array = array(scan_res, dtype=np.float).transpose()

            ylim = [nanmin(data_array[:, 1:]), nanmax(data_array[:, 1:]) * 1.1]
        # print data_array.shape
        if init_return:
            self.mod.SetQuiet()
            setattr(self.mod, parameter, init)
            self.mod.doMca()
            self.mod.SetLoud()

        mm = ModelMap(self.mod)
        species = mm.hasSpecies()
        if parameter in species:
            x_label = '[%s]' % parameter.replace('_', ' ')
        else:
            x_label = parameter
        ax_properties = {'ylabel': y_label,
                         'xlabel': x_label,
                         'xscale': 'linear',
                         'yscale': 'linear',
                         'xlim': [find_min(scan_range), find_max(scan_range)],
                         'ylim': ylim}

        data = Data2D(mod=self.mod,
                      column_names=column_names,
                      data_array=data_array,
                      ltxe=self._ltxe,
                      analysis_method='symca',
                      ax_properties=ax_properties,
                      file_name=self.name)

        return data

    def _calc_abs_value(self):
        """Calculates the absolute numeric value of the control coefficient from the
           values of its control patterns."""
        keys = self.expression.atoms(Symbol)
        subsdict = {}
        if len(keys) == 0:
            subsdict = None
        for key in keys:
            str_key = str(key)
            subsdict[str_key] = getattr(self.mod, str_key)
        for pattern in self.control_patterns.values():
            pattern._calc_value(subsdict)
        self._abs_value = sum(
            [abs(pattern._value) for pattern in self.control_patterns.values()])

    def _calc_value(self):
        """Calculates the numeric value of the control coefficient from the
           values of its control patterns."""
        keys = self.expression.atoms(Symbol)
        subsdict = {}
        if len(keys) == 0:
            subsdict = None
        for key in keys:
            str_key = str(key)
            subsdict[str_key] = getattr(self.mod, str_key)
        for pattern in self.control_patterns.values():
            pattern._calc_value(subsdict)
        self._value = sum(
            [pattern._value for pattern in self.control_patterns.values()])

    def _set_control_patterns(self):
        """Divides control coefficient into control patterns and saves
           results in self.CPx where x is a number is the number of the
           control pattern as it appears in in control coefficient
           expression"""
        patterns = self.numerator.as_coeff_add()[1]
        if len(patterns) == 0:
            patterns = [self.numerator.as_coeff_add()[0]]

        cps = DotDict()
        cps._make_repr('v.name', 'v.value', formatter_factory())
        for i, pattern in enumerate(patterns):
            name = 'CP{:3}'.format(i + 1).replace(' ', '0')
            cp = CPattern(self.mod,
                          name,
                          pattern,
                          self.denominator_object,
                          self,
                          self._ltxe)
            setattr(self, name, cp)
            cps[name] = cp
        self.control_patterns = cps
        # assert self._check_control_patterns == True

    def _check_control_patterns(self):
        """Checks that all control patterns are either positive or negative"""
        all_same = False
        poscomp = [i.value > 0 for i in self.control_patterns.values()]
        negcomp = [i.value < 0 for i in self.control_patterns.values()]
        if all(poscomp):
            all_same = True
        elif all(negcomp):
            all_same = True
        return all_same

    def highlight_patterns(self, width=None, height=None,
                           show_dummy_sinks=False,
                           show_external_modifier_links=False,
                           pos_dic=None):

        mg = ModelGraph(mod=self.mod, pos_dic=pos_dic,
                        analysis_method=self._analysis_method)
        if height:
            mg.height = height
        if width:
            mg.width = width

        mg.highlight_cc(self, show_dummy_sinks, show_external_modifier_links)


class CPattern(CCBase):

    """docstring for CPattern"""

    def __init__(self,
                 mod,
                 name,
                 expression,
                 denominator,
                 parent,
                 ltxe):
        super(CPattern, self).__init__(mod,
                                       name,
                                       expression,
                                       ltxe)
        self.numerator = expression
        self.denominator = denominator.expression
        self.expression = self.numerator / denominator.expression
        self.denominator_object = denominator
        self.parent = parent

        self._latex_numerator = None
        self._latex_expression_full = None
        self._latex_expression = None
        self._latex_name = None
        self._percentage = None

    def _calc_value(self, subsdict=None):
        """Calculates the value of the expression"""
        if subsdict is None:
            keys = self.expression.atoms(Symbol)
            subsdict = {}
            for key in keys:
                str_key = str(key)
                subsdict[str_key] = getattr(self.mod, str_key)
        self._value = get_value(self._str_expression, subsdict)

    @property
    def latex_numerator(self):
        if not self._latex_numerator:
            self._latex_numerator = self._ltxe.expression_to_latex(
                self.numerator
            )
        return self._latex_numerator

    @property
    def latex_expression_full(self):
        if not self._latex_expression_full:
            full_expr = '\\frac{' + self.latex_numerator + '}{' \
                + self.denominator_object.latex_expression + '}'
            self._latex_expression_full = full_expr
        return self._latex_expression_full

    @property
    def latex_expression(self):
        if not self._latex_expression:
            self._latex_expression = self.latex_numerator + '/~\\Sigma'
        return self._latex_expression

    @property
    def latex_name(self):
        if not self._latex_name:
            self._latex_name = self.name
        return self._latex_name

    @property
    def percentage(self):
        self._percentage = (abs(self.value) / self.parent.abs_value) * 100
        return self._percentage



"""NetworkX graphs do not have events that can be listened to.  In order to 
watch the NetworkX graph object for changes a custom eventful graph object must 
be created.  The custom eventful graph object will inherit from the base graph 
object and use special eventful dictionaries instead of standard Python dict 
instances.  Because NetworkX nests dictionaries inside dictionaries, it's 
important that the eventful dictionary is capable of recognizing when a 
dictionary value is set to another dictionary instance.  When this happens, the 
eventful dictionary needs to also make the new dictionary an eventful 
dictionary.  This allows the eventful dictionary to listen to changes made to 
dictionaries within dictionaries."""
import networkx
from networkx.generators.classic import empty_graph

from eventful_dict import EventfulDict

class EventfulGraph(networkx.Graph):

    _constructed_callback = None

    @staticmethod
    def on_constructed(callback):
        """Register a callback to be called when a graph is constructed."""
        if callback is None or callable(callback):
            EventfulGraph._constructed_callback = callback
    
    def __init__(self, *pargs, **kwargs):
        """Initialize a graph with edges, name, graph attributes.
        
        Parameters
        sleep: float
            optional float that allows you to tell the
        dictionary to hang for the given amount of seconds on each
        event.  This is usefull for animations."""
        super(EventfulGraph, self).__init__(*pargs, **kwargs)

        # Override internal dictionaries with custom eventful ones.
        sleep = kwargs.get('sleep', 0.0)
        self.graph = EventfulDict(self.graph, sleep=sleep)
        self.node = EventfulDict(self.node, sleep=sleep)
        self.adj = EventfulDict(self.adj, sleep=sleep)

        # Notify callback of construction event.
        if EventfulGraph._constructed_callback:
            EventfulGraph._constructed_callback(self)


def empty_eventfulgraph_hook(*pargs, **kwargs):
    def wrapped(*wpargs, **wkwargs):
        """Wrapper for networkx.generators.classic.empty_graph(...)"""
        wkwargs['create_using'] = EventfulGraph(*pargs, **kwargs)
        return empty_graph(*wpargs, **wkwargs)
    return wrapped



from . import misc
from . import plotting
from . import model_graph
from .config import ConfigReader
from .model_comparing import compare_models, SteadyStateComparer, SimulationComparer, ParameterScanComparer, ClosedOpenComparer


from os import path
from re import match, findall, sub

import pysces
from sympy import Symbol, sympify
from datetime import datetime



# File reading/validation functions
def get_term_types_from_raw_data(raw_data_dict):
    """
    Determines the types of terms defined for ThermoKin based on the
    file contents. This allows for generation of latex expressions
    based on these terms.

    Parameters
    ----------
    raw_data_dict : dict of str:{str:str}

    Returns
    -------
    set of str

    """
    term_types = set()
    for v in raw_data_dict.itervalues():
        for k in v.iterkeys():
            term_types.add(k)
    return term_types


class FormatException(Exception):
    pass


def read_reqn_file(path_to_file):
    """
    Reads the contents of a file and returns it as a list of lines.

    Parameters
    ----------
    path_to_file : str
        Path to file that is to read in

    Returns
    -------
    list of str
        The file contents as separate strings in a list

    """
    with open(path_to_file) as f:
        lines = f.readlines()
    return lines


def get_terms(raw_lines, term_type):
    """
    Takes a list of strings and returns a new list containing only
    lines starting with `term_type` and strips line endings.

    Term can be either of the "main" (or `!T`) type or additional (or
    `!G`) type

    Parameters
    ----------
    raw_lines : list of str
        List of lines from a '.reqn' file.
    term_type : str
        This string specifies the type of term.

    Returns
    -------
    list of str
    """
    assert term_type == '!T' or term_type == '!G', 'Invalid term type specified'
    valid_prefix_lines = [line for line in raw_lines if line.startswith(term_type)]
    no_line_endings = []
    for line in valid_prefix_lines:
        if line[-1] == '\n':
            no_line_endings.append(line[:-1])
        else:
            no_line_endings.append(line)
    return no_line_endings


def check_term_format(lines, term_type):
    """
    Inspects a list of string for the correct ThermoKin syntax. Returns
    `True` in case of correct format. Throws exception otherwise.

    Correct format is a str matching the pattern "X{\w*}{\w*} .*" . Where
    "X" is either "!G" or "!T" as specified by `term_type`.

    Parameters
    ----------
    lines : list of str
        Clean list of lines from a '.reqn' file.

    term_type : str
        This string specifies the type of term.

    Returns
    -------
    bool

    """
    assert term_type == '!T' or term_type == '!G', 'Invalid term type specified'
    errors_in = []
    for i, line in enumerate(lines):
        if not match(term_type + '{\w*}{\w*} .*', line):
            errors_in.append(str(i))
    if len(errors_in) == 0:
        return True
    else:
        error_str = ', '.join(errors_in)
        raise FormatException('Incorrect syntax in lines:' + error_str)


def construct_dict(lines):
    """
    Constructs a dictionary of dictionaries for each reaction.

    Here keys of the outer dictionary is reaction name strings while
    the inner dictionary keys are the term names. The inner dictionary
    values are the term expressions

    Parameters
    ----------
    lines : list of str

    Returns
    -------
    dict of str:{str:str}
    """
    outer_dict = {}
    for line in lines:
        in_brackets = findall('(?<={)\w+', line)
        r_name = in_brackets[0]
        t_name = in_brackets[1]
        expr = findall('(?<=\w} ).*', line)[0]

        inner_dict = {t_name: expr}
        if r_name in outer_dict:
            outer_dict[r_name].update(inner_dict)
        else:
            outer_dict[r_name] = inner_dict
    return outer_dict


def get_subs_dict(expression, mod):
    """
    Builds a substitution dictionary of an expression based of the
    values of these symbols in a model.

    Parameters
    ----------
    expression : sympy expression
    mod : PysMod

    Returns
    -------
    dict of sympy.Symbol:float

    """
    subs_dict = {}
    symbols = expression.atoms(Symbol)
    for symbol in symbols:
        attr = str(symbol)
        subs_dict[attr] = getattr(mod, attr)
    return subs_dict


def get_reqn_path(mod):
    """
    Gets the default path and filename of`.reqn` files belonging to a model

    The `.reqn` files which contain rate equations split into different
    (arbitrary) components should be saved in the same directory as the model
    file itself by default. It should have the same filename (sans extension)
    as the model file.

    Parameters
    ----------
    mod : PysMod
        A pysces model which has corresponding `.reqn` file saved in the same
        directory with the same file name as the model file.

    Returns
    -------
    str
        A sting with the path and filename of the `.reqn` file.
    """

    fname = mod.ModelFile
    dot_loc = fname.find('.')
    fname_min_ext = fname[:dot_loc]
    fname_ext = fname_min_ext + '.reqn'
    return path.join(mod.ModelDir, fname_ext)


def get_term_dict(raw_lines, term_type):
    """
    Returns the term dictionary from a list of raw lines from a file.

    The contents of a '.reqn' file is read and passed to this function.
    Here the contents is parsed and 'main terms' are extracted and
    returned as a dict of str:{str:str}.

    Parameters
    ----------
    raw_lines : list of str
        List of lines from a '.reqn' file.

    Returns
    -------
    dict of str:{str:str}
    """
    clean_terms = get_terms(raw_lines, term_type)
    if check_term_format(clean_terms, term_type):
        term_dict = construct_dict(clean_terms)
        return term_dict


def get_all_terms(path_to_read):
    raw_lines = read_reqn_file(path_to_read)
    main_terms = get_term_dict(raw_lines, '!T')
    add_terms = get_term_dict(raw_lines, '!G')
    return  main_terms, add_terms




# File writing/validation functions
def get_str_formulas(mod):
    """
    Returns a dictionary with reaction_name:string_formula as  key:value
    pairs.

    Goes through mod.reactions and constructs a dictionary where
    reaction_name is the key and mod.reaction_name.formula is  the
    value.

    Parameters
    ----------
    mod : PysMod
        The model which will be used to construct the dictionary

    Returns
    -------
    dict of str:str
        A dictionary with reaction_name:string_formula as
        key:value pairs

    """
    string_formulas = {}
    for reaction in mod.reactions:
        string_formulas[reaction] = getattr(mod, reaction).formula
    return string_formulas


def replace_pow(str_formulas):
    """
    Creates new dict from an existing dict with "pow(x,y)" in values
    replaced with "x**y".

    Goes through the values of an dictionary and uses regex to convert
    the pysces internal syntax for powers with  standard python syntax.
    This is needed before conversion to sympy expressions. This use case
    requires reaction  names as they appear in pysces as keys.

    Parameters
    ----------
    str_formulas : dict of str:str
        A dictionary where the values as contain pysces format strings
        representing rate equation expressions with powers in the syntax
        "pow(x,y)"

    Returns
    -------
    dict of str:str
        A new dictionary with str rate equations where powers are
        represented by standard python syntax e.g. x**y

    """
    new_str_formulas = {}
    for k, v in str_formulas.iteritems():
        new_str_formulas[k] = sub(r'pow\((\S*?),(\S*?)\)', r'\1**\2', v)
    return new_str_formulas


def get_sympy_formulas(str_formulas):
    """
    Converts dict with str values to sympy expression values.

    Used to convert key:string_formula to key:sympy_formula. Intended
    use case is for automatic separation of rate equation terms into
    mass action and binding terms. This  use case requires reaction
    names as they appear in  pysces as keys.

    Parameters
    ----------
    str_formulas : dict of str:str
        Dictionary with str values that represent reaction expressions.
        This dictionary needs to have already  passed through all
        sanitising functions/methods (e.g. `replace_pow`).

    Returns
    -------
    dict with sympy_expression values and original keys
        Dictionary where values are symbolic sympy expressions

    """
    return {k: sympify(v) for (k, v) in str_formulas.items()}


def get_sympy_terms(sympy_formulas):
    """
    Converts a dict with sympy expressions as values to a new dict with
    list values containing either the  original expression or a negative
    and a positive  expressions.

    This is used to separate reversible and irreversible  reactions.
    Reversible reactions will have two terms, one negative and one
    positive. Here expressions are expanded and split into terms and
    tested for the  above criteria: If met the dict value will be a list
    of two expressions, each representing a term of the  rate equation.
    Otherwise the dict value will be a list with a single item - the
    original expression. This  use case requires reaction names as they
    appear  in pysces as keys.

    Parameters
    ----------
    sympy_formulas : dict of str:sympy expression values
        Dictionary with values representing rate equations as sympy
        expressions. Keys are reaction names

    Returns
    -------
    dict of str:list sympy expression
        Each list will have either have one item, the original dict
        value OR two items -the original dict value split into a
        negative and positive expression.

    See Also
    --------
    check_for_negatives
    """
    sympy_terms = {}
    for name, formula in sympy_formulas.iteritems():
        terms = formula.expand().as_coeff_add()[1]
        if len(terms) == 2 and check_for_negatives(terms):
            sympy_terms[name] = terms
        else:
            sympy_terms[name] = [formula.factor()]
    return sympy_terms


def get_ma_terms(mod, sympy_terms):
    """
    Returns dict with reaction names as keys and mass action terms as
    values from a dict with reaction names as keys and lists of sympy
    expressions as values.

    Only reversible reactions are handled. Any list in the ``sympy_terms``
    dict that does not have a length of 2 will be ignored.

    Parameters
    ----------
    mod : PysMod
        The model from which the `sympy_terms` dict was originally
        constructed.
    sympy_terms: dict of str:list of sympy expressions
        This dictionary should be created by `get_sympy_terms`.

    Returns
    -------
    dict of str:sympy expression
        Each value will be a mass action term for each reaction key with
        a form depending on reversibility as described above.

    See Also
    --------
    get_st_pt_keq
    get_sympy_terms
    sort_terms
    """
    model_map = pysces.ModelMap(mod)  # model map to get substrates, products
    # and parameters for each reaction

    messages = {}
    ma_terms = {}
    for name, terms in sympy_terms.iteritems():
        reaction_map = getattr(model_map, name)

        substrates = [sympify(substrate) for substrate in
                      reaction_map.hasSubstrates()]

        products = [sympify(product) for product in reaction_map.hasProducts()]

        if len(terms) == 2:  # condition for reversible reactions
            # make sure negative term is second in term list
            terms = sort_terms(terms)
            # divide pos term by neg term and factorise
            expressions = (-terms[0] / terms[1]).factor()
            # get substrate, product and keq terms (and strategy)
            st, pt, keq, message = get_st_pt_keq(expressions, substrates,
                                                 products)
            if all([st, pt, keq]):
                ma_terms[name] = st - pt / keq
            messages[name] = message
        else:
            messages[
                name] = 'rate equation not included - irreversible or unknown form'
    return ma_terms, messages


def get_st_pt_keq(expression, substrates, products):
    """
    Takes an expression representing "substrates/products *
    Keq_expression" and returns substrates, products and keq_expression
    separately.


    Parameters
    ----------
    expression : sympy expression
        The expression containing "substrates/products * Keq_expression"
    substrates : list of sympy symbols
        List with symbolic representations for each substrate involved
        in the reaction which `expression` represents.
    products : list of sympy symbols
        List with symbolic representations for each product involved in
        the reaction which `expression` represents.
        Returns
    -------
    tuple of sympy expressions and int
        This tuple contains sympy expressions for the substrates,
        products and keq_expression in that order. The final value will
        be an int which indicates the strategy followed.

    See Also
    --------
    st_pt_keq_from_expression

    """

    res = st_pt_keq_from_expression(expression,
                                    substrates,
                                    products)
    subs_term, prod_term, keq, message = res

    return subs_term, prod_term, keq, message


def st_pt_keq_from_expression(expression, substrates, products,
                              failure_threshold=10):
    """
    Take an expression representing "substrates/products *
    Keq_expression" and returns substrates, products and keq_expression
    separately.

    In this strategy there is no inspection of the stoichiometry as
    provided by the model map. Here the expressions is
    divided/multiplied by each substrate/product until it no longer
    appears in the expression. If the substrates or products are not
    removed after a defined number of attempts a total failure occurs
    and the function returns `None`

    This is a fallback for cases where
    defined stoichiometry does not correspond to the actual rate
    equation.

    Here cases where the substrate/product do not appear in the rate
    equation at all throws an assertion error.

    Parameters
    ----------
    expression : sympy expression
        The expression containing "substrates/products * Keq_expression"
    substrates : list of sympy symbols
        List with symbolic representations for each substrate involved
        in the reaction which `expression` represents.
    products : list of sympy symbols
        List with symbolic representations for each product involved in
        the reaction which `expression` represents.
    failure_threshold : int, optional (Default: 10)
        A threshold value the defines the number of times the metabolite
        removal strategy should be tried before failure.

    Returns
    -------
    tuple of sympy_expressions or `None`
        This tuple contains sympy expressions for the substrates,
        products and keq_expression in that order. None is returned if
        this strategy fails.

    """
    new_expression = expression
    subs_term = 1
    prod_term = 1
    fail = False
    message = 'successful separation of rate equation terms'
    # Remove substrates from expression by division
    # Each division multiplies subs_term with substrate
    for substrate in substrates:
        # divide expr by subs while subs in expr
        if substrate not in new_expression.atoms(Symbol):
            fail = True
            message = 'failure: substrate %s not in rate equation' % str(
                substrate)
            break
        tries = 0
        while substrate in new_expression.atoms(Symbol):
            new_expression = new_expression / substrate
            subs_term *= substrate
            tries += 1
            if tries > failure_threshold:
                message = 'failure: cannot remove substrate %s from rate equation' % str(
                    substrate)
                fail = True
                break
        if fail:
            break
    # Same as above but for products
    # Product removed by multiplication
    if not fail:
        for product in products:
            if product not in new_expression.atoms(Symbol):
                fail = True
                message = 'failure: product %s not in rate equation' % str(
                    product)
                break
            tries = 0
            while product in new_expression.atoms(Symbol):
                new_expression = new_expression * product
                prod_term *= product
                tries += 1
                if tries > failure_threshold:
                    message = 'failure: cannot remove product %s from rate equation' % str(
                        product)
                    fail = True
                    break
            if fail:
                break
    keq = new_expression.subs({1.0: 1})
    if fail:
        return 0, 0, 0, message
    else:
        return subs_term, prod_term, keq, message


def sort_terms(terms):
    """
    Returns a list of two sympy expressions where the expression is
    positive and the second expression is negative.

    Parameters
    ----------
    terms : list of sympy expressions
        A list with length of 2 where one element is positive and
        the other is negative (starts with a minus symbol)

    Returns
    -------
    tuple of sympy expressions
        A tuple where the first element is positive and the
        second is negative.
    """

    neg = None
    pos = None
    for term in terms:
        if str(term)[0] == '-':  # negative terms should start with a '-'
            neg = term
        else:
            pos = term
    assert neg, 'No negative terms ' + str(terms)
    assert pos, 'No positive terms ' + str(terms)
    return pos, neg


def get_binding_vc_terms(sympy_formulas, ma_terms):
    """
    Returns dictionary with a combined "rate capacity" and "binding"
    term as values.

    Uses the symbolic rate equations dictionary and mass action term
    dictionaries to construct a new dictionary with "rate capacity-
    binding" terms. The symbolic rate equations are divided by their
    mass action terms. The results are the "rate capacity-binding"
    terms. This use case requires reaction names as they appear in
    pysces as keys for both dictionaries.

    Parameters
    ----------
    sympy_formulas : dict of str:sympy expression

        Full rate equations for all reactions in model. Keys are
        reaction names and correspond to this in `ma_terms`.

    ma_terms : dict of str:sympy expression

        Mass action terms for all reactions in model. Keys are reaction
        names and correspond to this in `sympy_formulas`.

    Returns
    -------
    dict of str:sympy expression
        A dictionary with reaction names as keys and sympy
        expressions representing "rate capacity-binding"
        terms as values.
    """
    binding_terms = {}
    for name, ma_term in ma_terms.iteritems():
        binding_terms[name] = (sympy_formulas[name] / ma_term).factor().factor()
    return binding_terms


def check_for_negatives(terms):
    """
    Returns `True` for a list of sympy expressions contains any
    expressions that are negative.

    Parameters
    ----------
    terms : list of sympy expressions
        A list where expressions may be either positive or  negative.

    Returns
    -------
    bool
        `True` if any negative terms in expression. Otherwise
        `False`

    """
    any_negs = False
    for term in terms:
        if str(term)[0] == '-':
            any_negs = True
    return any_negs


def create_reqn_data(mod):
    string_formulas = get_str_formulas(mod)
    string_formulas = replace_pow(string_formulas)
    sympy_formulas = get_sympy_formulas(string_formulas)
    sympy_terms = get_sympy_terms(sympy_formulas)
    non_irr = filter_irreversible(sympy_terms)
    gamma_keq_terms, _ = get_gamma_keq_terms(mod, non_irr)
    ma_terms, messages = get_ma_terms(mod, sympy_terms)
    binding_vc_terms = get_binding_vc_terms(sympy_formulas, ma_terms)
    return ma_terms, binding_vc_terms, gamma_keq_terms, messages


def create_gamma_keq_reqn_data(mod):
    string_formulas = get_str_formulas(mod)
    string_formulas = replace_pow(string_formulas)
    sympy_formulas = get_sympy_formulas(string_formulas)
    sympy_terms = get_sympy_terms(sympy_formulas)
    sympy_terms = filter_irreversible(sympy_terms)
    gamma_keq, messages = get_gamma_keq_terms(mod, sympy_terms)
    return gamma_keq, messages


def get_gamma_keq_terms(mod, sympy_terms):
    model_map = pysces.ModelMap(mod)  # model map to get substrates, products
    # and parameters for each reaction

    messages = {}
    gamma_keq_terms = {}
    for name, terms in sympy_terms.iteritems():
        reaction_map = getattr(model_map, name)

        substrates = [sympify(substrate) for substrate in
                      reaction_map.hasSubstrates()]

        products = [sympify(product) for product in reaction_map.hasProducts()]

        if len(terms) == 2:  # condition for reversible reactions
            # make sure negative term is second in term list
            terms = sort_terms(terms)
            # divide pos term by neg term and factorise
            expressions = (-terms[0] / terms[1]).factor()
            # get substrate, product and keq terms (and strategy)
            st, pt, keq, _ = get_st_pt_keq(expressions, substrates,
                                                 products)
            if all([st, pt, keq]):
                gamma_keq_terms[name] = pt / (keq*st)
                messages[name] = 'successful generation of gamma/keq term'
            else:
                messages[name] = 'generation of gamma/keq term failed'

    return gamma_keq_terms, messages


def filter_irreversible(sympy_terms):
    new_sympy_terms = {}
    for k, v in sympy_terms.iteritems():
        if len(v) == 2:
            new_sympy_terms[k] = v
    return new_sympy_terms


def write_reqn_file(file_name, model_name, ma_terms, vc_binding_terms, gamma_keq_terms, messages):
    already_written = []
    date = datetime.strftime(datetime.now(), '%H:%M:%S %d-%m-%Y')
    with open(file_name, 'w') as f:
        f.write('# Automatically parsed and split rate equations for model: %s\n' % model_name)
        f.write('# generated on: %s\n\n' % date)
        f.write('# Note that this is a best effort attempt that is highly dependent\n')
        f.write('# on the form of the rate equations as defined in the model file.\n')
        f.write('# Check correctness before use.\n\n')
        for reaction_name, ma_term in ma_terms.iteritems():
            already_written.append(reaction_name)
            f.write('# %s :%s\n' % (reaction_name, messages[reaction_name]))
            f.write('!T{%s}{ma} %s\n' % (reaction_name, ma_term))
            f.write('!T{%s}{bind_vc} %s\n' % (
                reaction_name, vc_binding_terms[reaction_name]))
            f.write('!G{%s}{gamma_keq} %s\n' % (reaction_name, gamma_keq_terms[reaction_name]))
            f.write('\n')
        for k, v in messages.iteritems():
            if k not in already_written:
                f.write('# %s :%s\n' % (k, v))


def term_to_file(file_name, expression, parent_name=None, term_name=None ):
    date = datetime.strftime(datetime.now(), '%H:%M:%S %d-%m-%Y')
    if not parent_name:
        parent_name = 'undefined'
    if not term_name:
        term_name = 'undefined'
    with open(file_name,'a') as f:
        f.write('\n')
        f.write('# Additional term appended on %s\n' % date)
        if 'undefined' in (term_name,parent_name):
            print 'Warning: writing partially defined term to %s. Please inspect file for further details.' % file_name
            f.write('# The term below is partially defined - fix term manually by defining reaction and term names\n')
        f.write('!G{%s}{%s} %s\n' % (parent_name,
                                     term_name,
                                     expression))





# There functions are not used anymore
#
# def get_gamma_keq_terms(mod, sympy_terms):
#     model_map = pysces.ModelMap(mod)  # model map to get substrates, products
#     # and parameters for each reaction
#
#     messages = {}
#     gamma_keq_terms = {}
#     for name, terms in sympy_terms.iteritems():
#         reaction_map = getattr(model_map, name)
#
#         substrates = [sympify(substrate) for substrate in
#                       reaction_map.hasSubstrates()]
#
#         products = [sympify(product) for product in reaction_map.hasProducts()]
#
#         if len(terms) == 2:  # condition for reversible reactions
#             # make sure negative term is second in term list
#             terms = sort_terms(terms)
#             # divide pos term by neg term and factorise
#             expressions = (-terms[0] / terms[1]).factor()
#             # get substrate, product and keq terms (and strategy)
#             st, pt, keq, _ = get_st_pt_keq(expressions, substrates,
#                                                  products)
#             if all([st, pt, keq]):
#                 gamma_keq_terms[name] = pt / (keq*st)
#                 messages[name] = 'successful generation of gamma/keq term'
#             else:
#                 messages[name] = 'generation of gamma/keq term failed'
#
#     return gamma_keq_terms, messages
#
# def create_gamma_keq_reqn_data(mod):
#     string_formulas = get_str_formulas(mod)
#     string_formulas = replace_pow(string_formulas)
#     sympy_formulas = get_sympy_formulas(string_formulas)
#     sympy_terms = get_sympy_terms(sympy_formulas)
#     sympy_terms = filter_irreversible(sympy_terms)
#     gamma_keq, messages = get_gamma_keq_terms(mod, sympy_terms)
#     return  gamma_keq, messages
#
# def get_irr_ma(expression, parameters, substrates, stoichiometry):
#     """
#     Returns a mass action expression for an irreversible reaction (which
#     simply consists of substrates).
#
#     Here two strategies are tried - if both fail, the answer from the
#     first strategy is used. For details refer to functions mentioned
#     under `See Also`.
#
#     Parameters
#     ----------
#     expression : sympy expression
#         A sympy expression representing a rate equation of an
#         irreversible reaction.
#     parameters : list of sympy symbols
#         List with symbolic representations for each parameter involved
#         in the reaction which `expression` represents.
#     substrates : list of sympy symbols
#         List with symbolic representations for each substrate involved
#         in the reaction which `expression` represents.
#     stoichiometry : dict of sympy.Symbol:float
#         Symbolic representations of the substrates and products are used
#         for the keys of this dict while the stoichiometric coefficient
#         values are floats.
#
#
#     Returns
#     -------
#     tuple of sympy expression and int
#         Symbolic expression for the mass action term of the irreversible
#         reaction and an integer indicating the strategy used.
#
#     See Also
#     --------
#     irr_ma_from_coeffs
#     irr_ma_from_expression
#     """
#     # strategy 1
#     strategy = 1
#     substrate_term = irr_ma_from_coeffs(substrates, stoichiometry)
#     valid = validate_irr_ma(expression, substrate_term)
#     if not valid:
#         # fallback strategy
#         strategy = 2
#         final_fallback = substrate_term
#         substrate_term = irr_ma_from_expression(expression, parameters)
#         # complete failure
#         if not substrate_term:
#             strategy = 3
#             substrate_term = final_fallback
#
#     return substrate_term, strategy
#
#
# def irr_ma_from_coeffs(substrates, stoichiometry):
#     """
#     Returns a mass action expression for an irreversible reaction (which
#     simply consists of substrates).
#
#     In this strategy the stoichiometric coefficients are used to
#     construct the substrate terms. Here an invalid substrate term  can be
#     produced when the rate equation does not follow  the stoichiometry
#     as defined in the model and the answer has to be validated using
#     `validate_irr_ma`.
#
#     Parameters
#     ----------
#     substrates : list of sympy symbols
#         List with symbolic representations for each substrate involved
#         in the reaction.
#     stoichiometry : dict of sympy.Symbol:float
#         Symbolic representations of the substrates and products are used
#         for the keys of this dict while the stoichiometric coefficient
#         values are floats.
#
#
#     Returns
#     -------
#     sympy expression
#         A symbolic expression of the substrate term of a mass action
#         expression for an irreversible reaction constructed using
#         stoichiometric coefficients.
#
#
#     """
#     return build_metabolite_term(substrates, stoichiometry)
#
#
# def irr_ma_from_expression(expression, parameters, failure_threshold=10):
#     """
#     Returns a mass action expression for an irreversible reaction (which
#     simply consists of substrates).
#
#     In this strategy there is no inspection of the stoichiometry as
#     provided by the model map. Here the expressions is divided or
#     multiplied by each parameter that initially appears in the
#     expression until it does not appear in the expression. If the
#     parameter is not removed after a defined number of attempts a total
#     failure occurs and this function returns `None`. This is a fallback
#     for cases where defined stoichiometry does not correspond to the
#     actual rate equation.
#
#     Parameters
#     ----------
#     expression : sympy expression
#         A sympy expression representing a rate equation of an
#         irreversible reaction.
#     parameters : list of sympy symbols
#         List with symbolic representations for each parameter involved
#         in the reaction which `expression` represents.
#     failure_threshold : int, optional (Default: 10)
#         A threshold value the defines the number of times the parameter
#         removal strategy should be tried before failure.
#
#     Returns
#     -------
#     sympy expression or None
#         A symbolic expression of the substrate term of a mass action
#         expression for an irreversible reaction constructed the rate
#         equation and parameters. None is returned in case of failure
#     """
#     expression_num = fraction(expression.expand())[0]
#     reset_point = expression_num
#     fail = False
#     for parameter in parameters:
#         tries = 0
#         switch_strat = False
#         while parameter in expression_num.atoms(Symbol):
#             expression_num = (expression_num / parameter).factor()
#             tries += 1
#             if tries > failure_threshold:
#                 switch_strat = True
#                 break
#
#         if switch_strat:
#             expression_num = reset_point
#             tries = 0
#             while parameter in expression_num.atoms(Symbol):
#                 expression_num = (expression_num * parameter).factor()
#                 tries += 1
#                 if tries > failure_threshold:
#                     fail = True
#                     break
#
#         if fail:
#             break
#         reset_point = expression_num
#     if fail:
#         return None
#     else:
#         return expression_num
#
#
# def validate_irr_ma(expression, substrate_term):
#     """
#     Returns `True` when the substrates in the substrates term has the same
#     number of coefficients as in the rate equation numerator.
#
#     In theory an expanded rate equation expression numerator of an
#     irreversible reaction should consist of only parameters and
#     substrates. Therefore, division of this numerator by the substrate
#     term should yield an expression without any substrates.
#
#     Parameters
#     ----------
#     expression : sympy expression
#         A sympy expression representing a rate equation of an
#         irreversible reaction
#     substrate_term : sympy expression
#         A sympy expression representing the substrate (mass action) term
#         of an irreversible reaction
#
#     Returns
#     -------
#     boolean
#         `True` for valid substrate term, otherwise `False`.
#     """
#     expression_num = fraction(expression.expand())[0]
#     remainder = expression_num / substrate_term
#     subs_atoms = substrate_term.atoms(Symbol)
#     valid = True
#     for remainder_atom in remainder.atoms(Symbol):
#         if remainder_atom in subs_atoms:
#             valid = False
#     return valid
#
# def build_metabolite_term(met_list, stoichiometry):
#     """
#     Given a list of metabolites and a dict with stoichiometry, this
#     function returns a metabolite term for a mass action expression.
#
#     Parameters
#     ----------
#     met_list : list of sympy.Symbol
#         List of symbolic representations of metabolites
#         (either products or substrates) that appear in a reaction.
#     stoichiometry : dict of sympy.Symbol:float
#         Symbolic representations of the metabolites are used as the keys
#         of this dict while the stoichiometric coefficients are floats
#
#     Returns
#     -------
#     sympy expression
#         A symbolic expression of the metabolite term of a mass action
#         expression constructed using stoichiometric coefficients.
#
#     See Also
#     --------
#     st_pt_keq_from_coeffs
#     """
#     met_term = 1
#     for met in met_list:
#         met_term *= met ** stoichiometry[met]
#
#     met_term = met_term.subs({1.0: 1})
#     return met_term
#
# def st_pt_keq_from_coeffs(expression, substrates, products, stoichiometry):
#     """
#     Takes an expression representing "substrates/products *
#     Keq_expression" and returns substrates, products and keq_expression
#     separately.
#
#     In this strategy the stoichiometric coefficients are used to
#     construct the substrate, product and Keq terms. Here an invalid Keq
#     expression can be produced when the rate equation does not follow
#     the stoichiometry as defined in the model and the answer has to be
#     validated using `validate_keq_expression`.
#
#     Parameters
#     ----------
#     expression : sympy expression
#         The expression containing "substrates/products * Keq_expression"
#     substrates : list of sympy symbols
#         List with symbolic representations for each substrate involved
#         in the reaction which `expression` represents.
#     products : list of sympy symbols
#         List with symbolic representations for each product involved in
#         the reaction which `expression` represents.
#     stoichiometry : dict of sympy.Symbol:float
#         Symbolic representations of the substrates and products are used
#         for the keys of this dict while the stoichiometric coefficients
#         are floats.
#
#     Returns
#     -------
#     tuple of sympy_expressions with length of 3
#         This tuple contains sympy expressions for the substrates,
#         products and keq_expression in that order
#
#     See Also
#     --------
#     get_st_pt_keq
#     st_pt_keq_from_expression
#     build_metabolite_term
#
#     """
#     subs_term = build_metabolite_term(substrates, stoichiometry)
#     prod_term = build_metabolite_term(products, stoichiometry)
#     keq = ((expression / subs_term) * prod_term).factor().subs({1.0: 1})
#     return subs_term, prod_term, keq
#
# def validate_keq_expression(expression, substrates, products):
#     """
#     Returns `True` when an expression does not contain any products
#     or substrates.
#
#     A valid Keq expression is either a single parameter representing the
#     Keq or it consists of parameters (maybe some variables) which
#     represents the Keq. There are no substrates or products of the
#     reaction in the Keq expression.
#
#     Parameters
#     ----------
#     expression : sympy expression
#         A symbolic expression representing the Keq. May be valid or
#         invalid.
#     substrates : list of sympy.Symbol
#         List of symbols for substrates involved in the reaction for
#         which `expression` is the Keq expression.
#     products : list of sympy.Symbol
#         List of symbols for products involved in the reaction for
#         which `expression` is the Keq expression.
#
#     Returns
#     -------
#     bool
#         True for valid Keq expression, False if invalid.
#
#     See Also
#     --------
#     st_pt_keq_from_coeffs
#     """
#     valid = True
#     expression_symbols = expression.atoms(Symbol)
#     for metabolite in substrates + products:
#         if metabolite in expression_symbols:
#             valid = False
#     return valid



import json
from os import path
from numpy import floor
from sympy import Symbol

from IPython.display import display
import ipywidgets as widgets
from pysces import ModelMap
from d3networkx_psctb import ForceDirectedGraph
from d3networkx_psctb import EventfulGraph

from ... import modeltools
from ..misc import split_coefficient


__author__ = 'carl'

__all__ = ['ModelGraph']


class ModelGraph(object):
    SPECIES_NODE = {'fill': "#FF6138",
                    'stroke': "#FF6138",
                    'color': 'black',
                    'shape': 'circle',
                    'dx': -12,
                    'dy': -15,
                    'font_size': 20,
                    'fixed': True,
                    'r': 11,
                    'strokewidth': '1.5px',
                    }

    REACTION_NODE = {'fill': "#00A388",
                     'stroke': "#00A388",
                     'color': 'black',
                     'shape': 'rect',
                     'dx': -12,
                     'dy': -15,
                     'font_size': 20,
                     'fixed': True,
                     'r': 8,
                     'strokewidth': '1.5px',
                     }

    RGB_RGB_RGB_ = {0: "rgb(94,79,162)", 1: "rgb(50,136,189)",
                    2: "rgb(102,194,165)", 3: "rgb(171,221,164)",
                    4: "rgb(230,245,152)", 5: "rgb(254,224,139)",
                    6: "rgb(253,174,97)", 7: "rgb(244,109,67)",
                    8: "rgb(213,62,79)", 9: "rgb(158,1,66)",
                    10: "rgb(158,1,66)"}

    DUMMY_SINK_NAMES = ['dummy','sink']


    def __init__(self, mod, pos_dic=None, analysis_method=None,
                 base_name=None):
        super(ModelGraph, self).__init__()
        self._eventful_graph = EventfulGraph(sleep=0.05)
        self._force_directed_graph = ForceDirectedGraph(self._eventful_graph)
        self._force_directed_graph.height = 500
        self._force_directed_graph.width = 900
        self.mod = mod
        self._model_map = ModelMap(mod)
        self._nodes_fixed = True
        if not analysis_method:
            analysis_method = 'modelgraph'
        self._analysis_method = analysis_method
        self._working_dir = modeltools.make_path(self.mod,
                                                 self._analysis_method)

        self._default_base_name = 'model_scheme'
        self._pod_dict_dir = modeltools.make_path(self.mod,
                                                  'modelgraph')

        if not base_name:
            base_name = self._default_base_name
        self._base_name = base_name

        self._top_box = widgets.VBox()
        self._big_box = widgets.VBox()
        self._big_box.children = (
            self._top_box, widgets.HTML('<hr>'), self._force_directed_graph)

        if not pos_dic:
            try:
                with open(path.join(self._pod_dict_dir,
                                    'graph_layout.dict'), 'r') as f:
                    self._pos_dic = json.load(f)
            except:
                self._clear_pos_dic()
        else:
            if pos_dic == 'clear':
                self._clear_pos_dic()
            else:
                with open(pos_dic, 'r') as f:
                    self._pos_dic = json.load(f)

        self._ec_dict = {}
        self._pos_change_setup()
        self._save_image_setup()

    def _clear_pos_dic(self):
        self._pos_dic = {}
        for sp in self._model_map.species:
            self._pos_dic[sp.name] = ('none', 'none')
        for rx in self._model_map.reactions:
            self._pos_dic[rx.name] = ('none', 'none')


    # TODO Change d3networkx_psctb so that dumping json and dumping svg work in the same way
    def _pos_change_setup(self):
        def make_pos_dic(nodes_json):
            raw_json = json.loads(nodes_json)
            pos_dic = {}
            for each in raw_json:
                pos_dic[str(each['id'])] = (each['x'], each['y'])
            return pos_dic

        def save_json(sender):
            pos_dic = make_pos_dic(self._force_directed_graph.gnode_json)
            self._pos_dic = pos_dic
            with open(path.join(self._pod_dict_dir, 'graph_layout.dict'),
                      'w') as f:
                json.dump(pos_dic, f)

        self._force_directed_graph.on_trait_change(save_json, 'gnode_json')

    def _save_image_setup(self):
        def save_image(sender, content):
            if 'button_click' in content:
                if content['button_click'] == 'save_image':
                    self.save()

        self._force_directed_graph.on_msg(save_image)


    def _make_species_nodes(self):
        for species in self._model_map.species:
            pos = self._pos_dic.get(species.name)
            if pos:
                x, y = pos
            else:
                x, y = "none", "none"
            self._eventful_graph.add_node(species.name,
                                          fill=ModelGraph.SPECIES_NODE['fill'],
                                          stroke=ModelGraph.SPECIES_NODE[
                                              'stroke'],
                                          color=ModelGraph.SPECIES_NODE[
                                              'color'],
                                          shape=ModelGraph.SPECIES_NODE[
                                              'shape'],
                                          label=species.name,
                                          dx=ModelGraph.SPECIES_NODE['dx'],
                                          dy=ModelGraph.SPECIES_NODE['dy'],
                                          font_size=ModelGraph.SPECIES_NODE[
                                              'font_size'],
                                          x=x,
                                          y=y,
                                          fixed=ModelGraph.SPECIES_NODE[
                                              'fixed'],
                                          r=ModelGraph.SPECIES_NODE['r'],
                                          )

    def _make_reaction_nodes(self):
        for reaction in self._model_map.reactions:
            pos = self._pos_dic.get(reaction.name)
            if pos:
                x, y = pos
            else:
                x, y = "none", "none"
            self._eventful_graph.add_node(reaction.name,
                                          fill=ModelGraph.REACTION_NODE[
                                              'fill'],
                                          stroke=ModelGraph.REACTION_NODE[
                                              'stroke'],
                                          color=ModelGraph.REACTION_NODE[
                                              'color'],
                                          shape=ModelGraph.REACTION_NODE[
                                              'shape'],
                                          label=reaction.name,
                                          dx=ModelGraph.REACTION_NODE['dx'],
                                          dy=ModelGraph.REACTION_NODE['dy'],
                                          font_size=ModelGraph.REACTION_NODE[
                                              'font_size'],
                                          x=x,
                                          y=y,
                                          fixed=ModelGraph.REACTION_NODE[
                                              'fixed'],
                                          r=ModelGraph.REACTION_NODE['r'],
                                          )

    def _make_substrate_links(self):
        for species in self._model_map.species:
            for reaction in species.isSubstrateOf():
                elas = 'ec%s_%s' % (reaction, species.name)
                self._eventful_graph.add_edge(reaction,
                                              species.name,
                                              name=elas,
                                              strokewidth="0px",
                                              distance=200,
                                              stroke='black')
                self._eventful_graph.adj[reaction][species.name][
                    'strokewidth'] = '2px'
                self._ec_dict[elas] = (reaction, species.name)
                del self._eventful_graph.adj[species.name][reaction]

    def _make_product_links(self):
        for species in self._model_map.species:
            for reaction in species.isProductOf():
                elas = 'ec%s_%s' % (reaction, species.name)
                self._eventful_graph.add_edge(species.name,
                                              reaction,
                                              name=elas,
                                              strokewidth="0px",
                                              distance=200,
                                              stroke='black',
                                              marker_end='arrowhead')
                self._eventful_graph.adj[reaction][species.name][
                    'strokewidth'] = '2px'
                del self._eventful_graph.adj[species.name][reaction]
                self._ec_dict[elas] = (reaction, species.name)

    def _make_modifier_links(self):
        for species in self._model_map.species:
            for reaction in species.isModifierOf():
                elas = 'ec%s_%s' % (reaction, species.name)
                self._eventful_graph.add_edge(reaction,
                                              species.name,
                                              name=elas,
                                              strokewidth="0px",
                                              distance=200,
                                              stroke='gray',
                                              marker_end='dagger', )
                self._eventful_graph.adj[species.name][reaction][
                    'strokewidth'] = '2px'
                del self._eventful_graph.adj[reaction][species.name]
                self._ec_dict[elas] = (species.name, reaction)


    def remove_external_modifier_links(self):
        for species in self._model_map.species:
            for reaction in species.isModifierOf():
                if species.name in self.mod.parameters:
                    self.change_link_properties(
                        'ec%s_%s' % (reaction, species.name),
                        {'strokewidth': '0px'})

    def remove_dummy_sinks(self):
        for each in ModelGraph.DUMMY_SINK_NAMES:
            if each in self._eventful_graph.node:
                del self._eventful_graph.node[each]

    def change_link_properties(self, elas, prop_dic=None,
                               only_overwrite=False):
        if not prop_dic:
            prop_dic = {}
        # TODO this can probably be replaced by misc.ec_dict(self.mod)[elas] or self._ec_dict can be constructed with that function.
        s_t = self._ec_dict.get(elas)
        if s_t:
            source, target = s_t
            dic_to_change = self._eventful_graph.adj[source][target]
            for k, v in prop_dic.iteritems():
                if only_overwrite and k in dic_to_change or not only_overwrite:
                    dic_to_change[k] = v

    def reset_node_properties(self):
        for species in self._model_map.species:
            self.change_node_properties(species.name, ModelGraph.SPECIES_NODE)

        for reaction in self._model_map.reactions:
            self.change_node_properties(reaction.name,
                                        ModelGraph.REACTION_NODE)


    def change_node_properties(self, node_name, prop_dic=None):
        if not prop_dic:
            prop_dic = {}
        dic = self._eventful_graph.node.get(node_name)
        if dic:
            for k, v in prop_dic.iteritems():
                dic[k] = v

    def draw_all_links(self):
        self._make_substrate_links()
        self._make_product_links()
        self._make_modifier_links()

    @property
    def nodes_fixed(self):
        return self._nodes_fixed

    @nodes_fixed.setter
    def nodes_fixed(self, value):
        self._nodes_fixed = value
        for node in self._model_map.reactions + self._model_map.species:
            self.change_node_properties(node.name,
                                        {'fixed': self._nodes_fixed})


    @property
    def straight_links(self):
        return self._force_directed_graph.straight_links

    @straight_links.setter
    def straight_links(self, value):
        self._force_directed_graph.straight_links = value

    @property
    def height(self):
        return self._force_directed_graph.height

    @height.setter
    def height(self, value):
        self._force_directed_graph.height = value

    @property
    def width(self):
        return self._force_directed_graph.width

    @width.setter
    def width(self, value):
        self._force_directed_graph.width = value

    def save(self, file_name=None):
        """Saves the image.

        Saves the image to either the default working directory or to an a
        specified location.
        Parameters
        ----------
        file_name : str, Optional (default : None)
            An optional path to which the image will be saved.

        Returns
        -------
        None
        """

        svg = self._force_directed_graph.svg_image

        file_name = modeltools.get_file_path(working_dir=self._working_dir,
                                             internal_filename=self._base_name,
                                             fmt='svg',
                                             file_name=file_name)
        with open(file_name, 'w') as f:
            f.write(svg)


    def show(self, no_links=False, clear_top_box=True):
        if clear_top_box:
            self._top_box.children = ()
        self._base_name = self._default_base_name
        display(self._big_box)
        self._eventful_graph.node.clear()
        self._make_species_nodes()
        self._make_reaction_nodes()
        self.nodes_fixed = self._nodes_fixed
        if not no_links:
            self.draw_all_links()


    def highlight_cp(self, cp, show_dummy_sinks=False,
                     show_external_modifier_links=False):
        colors = ModelGraph.RGB_RGB_RGB_
        symbols = cp.numerator.atoms(Symbol)
        title = widgets.Label(
            'Control Patterns for $' + cp.parent.latex_name + '$')
        CP_repr = widgets.Label('$~$')
        _, controlled, controller = split_coefficient(cp.parent.name, self.mod)
        species_names = [species.name for species in self._model_map.species]

        cat = floor(cp.percentage / 10)

        CP_repr.value = cp._repr_latex_()

        hr = widgets.HTML('<hr>')

        titles = widgets.VBox()
        titles.children = (title, hr, CP_repr, hr)
        self._top_box.children = (titles,)
        self.show(clear_top_box=False)
        self._make_product_links()
        self._make_substrate_links()
        self._make_modifier_links()
        self._force_directed_graph.show_color_legend = False
        self._force_directed_graph.show_color_legend = True
        self.reset_node_properties()
        if controlled in species_names:
            self.change_node_properties(controlled, {'stroke': '#8B361E',
                                                     'strokewidth': '3px'})
        self.change_node_properties(controller, {'stroke': '#006B58',
                                                 'strokewidth': '3px'})
        for sym in symbols:
            self.change_link_properties(str(sym),
                                        {'strokewidth': '20px',
                                         'marker_end': 'arrowheadlarge',
                                         'stroke': colors[cat]},
                                        only_overwrite=True)

        self._base_name = cp.parent.name + '_' + cp.name
        if not show_dummy_sinks:
            self.remove_dummy_sinks()
        if not show_external_modifier_links:
            self.remove_external_modifier_links()


    def highlight_cc(self, cc, show_dummy_sinks=False,
                     show_external_modifier_links=False):

        colors = ModelGraph.RGB_RGB_RGB_
        title = widgets.Label('Control Patterns for $' + cc.latex_name + '$')
        CP_repr = widgets.Label('$~$')
        _, controlled, controller = split_coefficient(cc.name, self.mod)
        species_names = [species.name for species in self._model_map.species]

        def onclick_maker(cp):
            symbols = cp.numerator.atoms(Symbol)

            def click(sender):
                self.reset_node_properties()
                self._base_name = cc.name + '_' + cp.name
                cat = floor(cp.percentage / 10)
                self._make_product_links()
                self._make_substrate_links()
                self._make_modifier_links()
                self._force_directed_graph.show_color_legend = False
                self._force_directed_graph.show_color_legend = True

                if controlled in species_names:
                    self.change_node_properties(controlled,
                                                {'stroke': '#8B361E',
                                                 'strokewidth': '3px'})
                self.change_node_properties(controller, {'stroke': '#006B58',
                                                         'strokewidth': '3px'})
                for sym in symbols:
                    self.change_link_properties(str(sym),
                                                {'strokewidth': '20px',
                                                 'marker_end': 'arrowheadlarge',
                                                 'stroke': colors[cat]},
                                                only_overwrite=True)

                CP_repr.value = cp._repr_latex_()
                if not show_dummy_sinks:
                    self.remove_dummy_sinks()
                if not show_external_modifier_links:
                    self.remove_external_modifier_links()

            return click

        button_list = []
        for k, v in cc.control_patterns.iteritems():
            but = widgets.Button()
            but.description = k
            click = onclick_maker(v)
            but.on_click(click)
            button_list.append(but)

        hr = widgets.HTML('<hr>')

        titles = widgets.VBox()
        titles.children = (title, hr, CP_repr, hr)
        buttons = widgets.HBox()
        buttons.children = tuple(button_list)
        self._top_box.children = (titles, buttons)
        self.show(clear_top_box=False)
        if not show_dummy_sinks:
            self.remove_dummy_sinks()
        if not show_external_modifier_links:
            self.remove_external_modifier_links()


from os import path
from colorsys import hsv_to_rgb, rgb_to_hsv
from collections import OrderedDict
from random import shuffle

import numpy
from pysces.PyscesModelMap import ModelMap
from pysces import Scanner
import pysces
from matplotlib.pyplot import get_cmap

from .. import modeltools
from ..latextools import LatexExpr
from ..utils.plotting import ScanFig, LineData, Data2D
from ..utils.misc import silence_print
from ..utils.misc import DotDict
from ..utils.misc import formatter_factory


exportLAWH = silence_print(pysces.write.exportLabelledArrayWithHeader)

__all__ = ['RateChar']


def strip_nan_from_scan(array_like):
    # this function assumes that column
    # zero contains valid data (the scan input)
    t_f = list(numpy.isnan(array_like[:, 1]))
    start = t_f.index(False)
    end = len(t_f) - t_f[::-1].index(False)

    return array_like[start:end, :]


class RateChar(object):
    def __init__(self, mod, min_concrange_factor=100,
                 max_concrange_factor=100,
                 scan_points=256,
                 auto_load=False):
        super(RateChar, self).__init__()

        self.mod = mod
        self.mod.SetQuiet()
        self._model_map = ModelMap(mod)
        self.mod.doState()

        self._analysis_method = 'ratechar'
        self._working_dir = modeltools.make_path(self.mod,
                                                 self._analysis_method)

        self._min_concrange_factor = min_concrange_factor
        self._max_concrange_factor = max_concrange_factor
        self._scan_points = scan_points

        self._ltxe = LatexExpr(self.mod)
        for species in self.mod.species:
            setattr(self, species, None)
        if auto_load:
            self.load_session()

    def do_ratechar(self, fixed='all',
                    scan_min=None,
                    scan_max=None,
                    min_concrange_factor=None,
                    max_concrange_factor=None,
                    scan_points=None,
                    solver=0,
                    auto_save=False):

        # this function wraps _do_scan functionality in a user friendly bubble
        if fixed == 'all':
            to_scan = self.mod.species
        elif type(fixed) is list or type(fixed) is tuple:
            for each in fixed:
                assert each in self.mod.species, 'Invalid species'
            to_scan = fixed
        else:
            assert fixed in self.mod.species, 'Invalid species'
            to_scan = [fixed]

        for each in to_scan:
            fixed_mod, fixed_ss = self._fix_at_ss(each)

            scan_start = self._min_max_chooser(fixed_ss,
                                               scan_min,
                                               min_concrange_factor,
                                               'min')

            scan_end = self._min_max_chooser(fixed_ss,
                                             scan_max,
                                             max_concrange_factor,
                                             'max')
            # here there could be a situation where a scan_min > scan_max
            # I wonder what will happen....

            if not scan_points:
                scan_points = self._scan_points

            column_names, results = self._do_scan(fixed_mod,
                                                  each,
                                                  scan_start,
                                                  scan_end,
                                                  scan_points)

            cleaned_results = strip_nan_from_scan(results)

            rcd = RateCharData(fixed_ss,
                               fixed_mod,
                               self.mod,
                               column_names,
                               cleaned_results,
                               self._model_map,
                               self._ltxe)
            setattr(self, each, rcd)
        if auto_save:
            self.save_session()

    def _min_max_chooser(self, ss, point, concrange, min_max):
        # chooses a minimum or maximum point based
        # on the information given by a user
        # ie if a specific min/max point is given - use that
        # if only concentration range is given -use that
        # if nothing is given - use the defualt conc_range_factor
        # pretty simple stuff
        if point:
            the_point = point
        if not point and concrange:
            if min_max == 'min':
                the_point = ss / concrange
            elif min_max == 'max':
                the_point = ss * concrange
        if not point and not concrange:
            if min_max == 'min':
                the_point = ss / self._min_concrange_factor
            elif min_max == 'max':
                the_point = ss * self._max_concrange_factor

        return the_point

    @silence_print
    def _do_scan(self,
                 fixed_mod,
                 fixed,
                 scan_min,
                 scan_max,
                 scan_points,
                 solver=0):
        # do scan is a simplified interface to pysces.Scanner
        # more intuitive than Scan1 (functional vs OO??)
        # returns the names of the scanned blocks together with
        # the results of the scan

        assert solver in (0, 1, 2), 'Solver mode can only be one of 0, 1 or 2'

        fixed_mod.mode_solver = solver

        demand_blocks = [
            'J_' + r for r in getattr(self._model_map, fixed).isSubstrateOf()]
        supply_blocks = [
            'J_' + r for r in getattr(self._model_map, fixed).isProductOf()]
        user_output = [fixed] + demand_blocks + supply_blocks

        scanner = Scanner(fixed_mod)
        scanner.quietRun = True
        scanner.addScanParameter(
            fixed, scan_min, scan_max, scan_points, log=True)
        scanner.addUserOutput(*user_output)
        scanner.Run()
        return user_output, scanner.UserOutputResults

    @silence_print
    def _fix_at_ss(self, fixed):
        # fixes the metabolite at the steady_state
        # (calls psctb.modeltools.fix_metabolite)
        # and returns both the ss value and the fixed model
        self.mod.doState()
        fixed_ss = getattr(self.mod, fixed + '_ss')
        fixed_mod = modeltools.fix_metabolite(self.mod, fixed)
        fixed_mod.SetQuiet()
        # i don't like this approach at all, too many possible unintended side
        # effects
        # setattr(fixed_mod, fixed, fixed_ss)
        # setattr(fixed_mod, 'fixed', fixed)
        # setattr(fixed_mod, 'fixed_ss', fixed_ss)
        fixed_mod.doState()
        return fixed_mod, fixed_ss

    def save_session(self, file_name=None):
        file_name = modeltools.get_file_path(working_dir=self._working_dir,
                                             internal_filename='save_data',
                                             fmt='npz',
                                             file_name=file_name,
                                             write_suffix=False)

        to_save = {}
        for species in self.mod.species:
            species_object = getattr(self, species)
            try:
                column_array = numpy.array(species_object._column_names)
                scan_results = species_object._scan_results

                to_save['col_{0}'.format(species)] = column_array
                to_save['res_{0}'.format(species)] = scan_results
            except:
                pass
        numpy.savez(file_name, **to_save)


    def save_results(self, folder=None, separator=',',format='%f'):
        base_folder = folder
        for species in self.mod.species:
            if folder:
                folder = path.join(base_folder, species)
            getattr(self, species).save_all_results(folder=folder,
                                                    separator=separator)

    def load_session(self, file_name=None):
        file_name = modeltools.get_file_path(working_dir=self._working_dir,
                                             internal_filename='save_data',
                                             fmt='npz',
                                             file_name=file_name,
                                             write_suffix=False)

        loaded_data = {}
        try:
            with numpy.load(file_name) as data_file:
                for k, v in data_file.iteritems():
                    loaded_data[k] = v
        except IOError as e:
            raise e

        for species in self.mod.species:
            try:
                column_names = [str(each) for each in
                                list(loaded_data['col_{0}'.format(species)])]
                scan_results = loaded_data['res_{0}'.format(species)]
                fixed_species = species
                fixed_mod, fixed_ss = self._fix_at_ss(fixed_species)
                rcd = RateCharData(fixed_ss=fixed_ss,
                                   fixed_mod=fixed_mod,
                                   basemod=self.mod, column_names=column_names,
                                   scan_results=scan_results,
                                   model_map=self._model_map, ltxe=self._ltxe)
                setattr(self, fixed_species, rcd)
            except:
                pass


class RateCharData(object):
    def __init__(self,
                 fixed_ss,
                 fixed_mod,
                 basemod,
                 column_names,
                 scan_results,
                 model_map,
                 ltxe):

        super(RateCharData, self).__init__()
        self.mod = fixed_mod

        self.scan_results = DotDict()
        self.mca_results = DotDict()

        self._slope_range_factor = 3.0

        self.scan_results['fixed'] = column_names[0]
        self.scan_results['fixed_ss'] = fixed_ss

        self.scan_results['scan_range'] = scan_results[:, 0]
        self.scan_results['flux_names'] = column_names[1:]
        self.scan_results['flux_data'] = scan_results[:, 1:]
        self.scan_results['scan_points'] = len(self.scan_results.scan_range)
        self.scan_results['flux_max'] = None
        self.scan_results['flux_min'] = None
        self.scan_results['scan_max'] = None
        self.scan_results['scan_min'] = None
        self.scan_results['ec_names'] = None
        self.scan_results['ec_data'] = None
        self.scan_results['rc_names'] = None
        self.scan_results['rc_data'] = None
        self.scan_results['prc_names'] = None
        self.scan_results['prc_data'] = None

        self._column_names = column_names
        self._scan_results = scan_results
        self._model_map = model_map

        self._analysis_method = 'ratechar'
        self._basemod = basemod
        self._working_dir = modeltools.make_path(self._basemod,
                                                 self._analysis_method,
                                                 [self.scan_results.fixed])
        self._ltxe = ltxe

        self._color_dict_ = None
        self._data_setup()
        self.mca_results._ltxe = ltxe
        self.mca_results._make_repr(
            '"$" + self._ltxe.expression_to_latex(k) + "$"', 'v',
            formatter_factory())
        # del self.scan_results
        # del self.mca_results

    def _data_setup(self):
        # reset value to do mcarc
        setattr(self.mod, self.scan_results.fixed, self.scan_results.fixed_ss)
        self.mod.doMcaRC()
        self._make_attach_total_fluxes()
        self._min_max_setup()
        self._attach_fluxes_to_self()
        self._make_all_coefficient_lines()
        self._attach_all_coefficients_to_self()
        self._make_all_summary()
        self._make_all_line_data()

    def _change_colour_order(self, order=None):
        if not order:
            order = self._color_dict_.keys()
            shuffle(order)
        self._color_dict_ = dict(zip(order, self._color_dict_.values()))
        self._make_all_line_data()

    def _make_all_line_data(self):
        self._make_flux_ld()
        self._make_ec_ld()
        self._make_rc_ld()
        self._make_prc_ld()
        self._make_total_flux_ld()

        self._line_data_dict = OrderedDict()
        self._line_data_dict.update(self._prc_ld_dict)
        self._line_data_dict.update(self._flux_ld_dict)
        self._line_data_dict.update(self._total_flux_ld_dict)

        self._line_data_dict.update(self._ec_ld_dict)
        self._line_data_dict.update(self._rc_ld_dict)

        del self._flux_ld_dict
        del self._ec_ld_dict
        del self._rc_ld_dict
        del self._prc_ld_dict
        del self._total_flux_ld_dict

    def _make_all_summary(self):
        self._make_ec_summary()
        self._make_cc_summary()
        self._make_rc_summary()
        self._make_prc_summary()

        self.mca_results.update(self._ec_summary)
        self.mca_results.update(self._cc_summary)
        self.mca_results.update(self._rc_summary)
        self.mca_results.update(self._prc_summary)

        del self._ec_summary
        del self._cc_summary
        del self._rc_summary
        del self._prc_summary

    def _make_ec_summary(self):
        ecs = {}
        reagent_of = [each[2:] for each in self.scan_results.flux_names]
        modifier_of = getattr(
            self._model_map, self.scan_results.fixed).isModifierOf()
        all_reactions = reagent_of + modifier_of
        for reaction in all_reactions:
            name = 'ec%s_%s' % (reaction, self.scan_results.fixed)
            val = getattr(self.mod, name)
            ecs[name] = val

        self._ec_summary = ecs

    def _make_rc_summary(self):
        rcs = {}
        for flux in self.scan_results.flux_names:
            reaction = flux[2:]
            name = '%s_%s' % (reaction, self.scan_results.fixed)
            val = getattr(self.mod.rc, name)
            name = 'rcJ' + name
            rcs[name] = val

        self._rc_summary = rcs

    def _make_cc_summary(self):

        ccs = {}
        reagent_of = [each[2:] for each in self.scan_results.flux_names]
        modifier_of = getattr(
            self._model_map, self.scan_results.fixed).isModifierOf()
        all_reactions = reagent_of + modifier_of

        for flux_reaction in reagent_of:
            for reaction in all_reactions:
                name = 'ccJ%s_%s' % (flux_reaction, reaction)
                val = getattr(self.mod, name)
                ccs[name] = val

        self._cc_summary = ccs

    def _make_prc_summary(self):

        prcs = {}

        reagent_of = [each[2:] for each in self.scan_results.flux_names]
        modifier_of = getattr(
            self._model_map, self.scan_results.fixed).isModifierOf()
        all_reactions = reagent_of + modifier_of

        for flux_reaction in reagent_of:
            for route_reaction in all_reactions:
                ec = getattr(self.mod,
                             'ec%s_%s' % (
                                 route_reaction, self.scan_results.fixed))

                cc = getattr(self.mod,
                             'ccJ%s_%s' % (flux_reaction, route_reaction))
                val = ec * cc
                name = 'prcJ%s_%s_%s' % (flux_reaction,
                                         self.scan_results.fixed,
                                         route_reaction)

                prcs[name] = val

        self._prc_summary = prcs

    def save_summary(self, file_name=None, separator=',',fmt='%f'):
        file_name = modeltools.get_file_path(working_dir=self._working_dir,
                                             internal_filename='mca_summary',
                                             fmt='csv',
                                             file_name=file_name, )

        keys = self.mca_results.keys()
        keys.sort()
        values = numpy.array([self.mca_results[k]
                              for k in keys]).reshape(len(keys), 1)

        try:
            exportLAWH(values,
                       names=keys,
                       header=['Value'],
                       fname=file_name,
                       sep=separator,
                       format=fmt)
        except IOError as e:
            print e.strerror

    def save_flux_results(self, file_name=None, separator=',',fmt='%f'):
        file_name = modeltools.get_file_path(working_dir=self._working_dir,
                                             internal_filename='flux_results',
                                             fmt='csv',
                                             file_name=file_name, )

        scan_points = self.scan_results.scan_points
        all_cols = numpy.hstack([
            self._scan_results,
            self.scan_results.total_supply.reshape(scan_points, 1),
            self.scan_results.total_demand.reshape(scan_points, 1)])
        column_names = self._column_names + ['Total Supply', 'Total Demand']

        try:
            exportLAWH(all_cols,
                       names=None,
                       header=column_names,
                       fname=file_name,
                       sep=separator,
                       format=fmt)
        except IOError as e:
            print e.strerror

    def save_coefficient_results(self,
                                 coefficient,
                                 file_name=None,
                                 separator=',',
                                 folder=None,
                                 fmt='%f'):
        assert_message = 'coefficient must be one of "ec", "rc" or "prc"'

        assert coefficient in ['rc', 'ec', 'prc'], assert_message

        base_name = coefficient + '_results'
        file_name = modeltools.get_file_path(working_dir=self._working_dir,
                                             internal_filename=base_name,
                                             fmt='csv',
                                             file_name=file_name, )

        results = getattr(self.scan_results, coefficient + '_data')
        names = getattr(self.scan_results, coefficient + '_names')
        new_names = []
        for each in names:
            new_names.append('x_vals')
            new_names.append(each)

        try:
            exportLAWH(results,
                       names=None,
                       header=new_names,
                       fname=file_name,
                       sep=separator,
                       format=fmt)
        except IOError as e:
            print e.strerror

    # TODO fix this method so that folder is a parameter only her
    def save_all_results(self, folder=None, separator=',',fmt='%f'):
        if not folder:
            folder = self._working_dir

        file_name = modeltools.get_file_path(working_dir=folder,
                                             internal_filename='flux_results',
                                             fmt='csv')
        self.save_flux_results(separator=separator, file_name=file_name,fmt=fmt)

        file_name = modeltools.get_file_path(working_dir=folder,
                                             internal_filename='mca_summary',
                                             fmt='csv')
        self.save_summary(separator=separator, file_name=file_name, fmt=fmt)
        for each in ['ec', 'rc', 'prc']:
            base_name = each + '_results'
            file_name = modeltools.get_file_path(working_dir=folder,
                                                 internal_filename=base_name,
                                                 fmt='csv')
            self.save_coefficient_results(coefficient=each,
                                          separator=separator,
                                          file_name=file_name,
                                          fmt=fmt)

    def _min_max_setup(self):
        # Negative minimum linear values mean nothing
        # because they don't translate to a log space
        # therefore we want the minimum non-negative/non-zero values.

        # lets make sure there are no zeros
        n_z_f = self.scan_results.flux_data[
            numpy.nonzero(self.scan_results.flux_data)]
        n_z_s = self.scan_results.scan_range[
            numpy.nonzero(self.scan_results.scan_range)]

        totals = numpy.vstack([self.scan_results.total_demand,
                               self.scan_results.total_supply])
        n_z_t = totals[numpy.nonzero(totals)]
        # and that the array is not now somehow empty
        # although if this happens-you have bigger problems
        if len(n_z_f) == 0:
            n_z_f = numpy.array([0.01, 1])
        if len(n_z_s) == 0:
            n_z_s = numpy.array([0.01, 1])

        # lets also (clumsily) find the non-negative mins and maxes
        # by converting to logspace (to get NaNs) and back
        # and then getting the min/max non-NaN
        # PS flux max is the max of the totals

        with numpy.errstate(all='ignore'):

            self.scan_results.flux_max = numpy.nanmax(10 ** numpy.log10(n_z_t))
            self.scan_results.flux_min = numpy.nanmin(10 ** numpy.log10(n_z_f))
            self.scan_results.scan_max = numpy.nanmax(10 ** numpy.log10(n_z_s))
            self.scan_results.scan_min = numpy.nanmin(10 ** numpy.log10(n_z_s))

    def _attach_fluxes_to_self(self):
        for i, each in enumerate(self.scan_results.flux_names):
            # setattr(self, each, self.scan_results.flux_data[:, i])
            self.scan_results[each] = self.scan_results.flux_data[:, i]

    def _attach_all_coefficients_to_self(self):
        setup_for = ['ec', 'rc', 'prc']
        for each in setup_for:
            eval('self._attach_coefficients_to_self(self.scan_results.' + each + '_names,\
                                                self.scan_results.' + each + '_data)')

    def _make_all_coefficient_lines(self):
        setup_for = ['ec', 'rc', 'prc']
        for each in setup_for:
            eval('self._make_' + each + '_lines()')

    def _make_attach_total_fluxes(self):
        demand_blocks = getattr(
            self._model_map, self.scan_results.fixed).isSubstrateOf()
        supply_blocks = getattr(
            self._model_map, self.scan_results.fixed).isProductOf()

        dem_pos = [self.scan_results.flux_names.index('J_' + flux)
                   for flux in demand_blocks]
        sup_pos = [self.scan_results.flux_names.index('J_' + flux)
                   for flux in supply_blocks]

        self.scan_results['total_demand'] = numpy.sum(
            [self.scan_results.flux_data[:, i]
             for i in dem_pos],
            axis=0)
        self.scan_results['total_supply'] = numpy.sum(
            [self.scan_results.flux_data[:, i]
             for i in sup_pos],
            axis=0)

    def _make_rc_lines(self):
        names = []
        resps = []

        for each in self.scan_results.flux_names:
            reaction = each[2:]
            name = reaction + '_' + self.scan_results.fixed

            J_ss = getattr(self.mod, each)
            slope = getattr(self.mod.rc, name)
            resp = self._tangent_line(J_ss, slope)

            name = 'rcJ' + name
            names.append(name)
            resps.append(resp)

        resps = numpy.hstack(resps)

        self.scan_results.rc_names = names
        self.scan_results.rc_data = resps

    def _make_prc_lines(self):
        names = []
        prcs = []

        reagent_of = [each[2:] for each in self.scan_results.flux_names]
        all_reactions = reagent_of + \
                        getattr(self._model_map,
                                self.scan_results.fixed).isModifierOf()

        for flux_reaction in self.scan_results.flux_names:

            J_ss = getattr(self.mod, flux_reaction)
            reaction = flux_reaction[2:]

            for route_reaction in all_reactions:
                ec = getattr(
                    self.mod,
                    'ec' + route_reaction + '_' + self.scan_results.fixed)
                cc = getattr(self.mod, 'ccJ' + reaction + '_' + route_reaction)
                slope = ec * cc

                prc = self._tangent_line(J_ss, slope)
                name = 'prcJ%s_%s_%s' % (reaction,
                                         self.scan_results.fixed,
                                         route_reaction)

                names.append(name)
                prcs.append(prc)

        prcs = numpy.hstack(prcs)

        self.scan_results.prc_names = names
        self.scan_results.prc_data = prcs

    def _make_ec_lines(self):
        names = []
        elasts = []

        for each in self.scan_results.flux_names:
            reaction = each[2:]
            name = 'ec' + reaction + '_' + self.scan_results.fixed

            J_ss = getattr(self.mod, each)
            slope = getattr(self.mod, name)
            elast = self._tangent_line(J_ss, slope)

            names.append(name)
            elasts.append(elast)

        elasts = numpy.hstack(elasts)

        self.scan_results.ec_names = names
        self.scan_results.ec_data = elasts

    def _attach_coefficients_to_self(self, names, tangent_lines):
        sp = 0
        ep = 2
        for name in names:
            # setattr(self, name, tangent_lines[:, sp:ep])
            self.scan_results[name] = tangent_lines[:, sp:ep]
            sp = ep
            ep += 2

    def _tangent_line(self, J_ss, slope):

        fix_ss = self.scan_results.fixed_ss

        constant = J_ss / (fix_ss ** slope)

        ydist = numpy.log10(self.scan_results.flux_max / self.scan_results.flux_min)
        xdist = numpy.log10(self.scan_results.scan_max / self.scan_results.scan_min)
        golden_ratio = (1 + numpy.sqrt(5)) / 2
        xyscale = xdist / (ydist * golden_ratio * 1.5)

        scale_factor = numpy.cos(numpy.arctan(slope * xyscale))
        distance = numpy.log10(self._slope_range_factor) * scale_factor

        range_min = fix_ss / (10 ** distance)
        range_max = fix_ss * (10 ** distance)
        scan_range = numpy.linspace(range_min, range_max, num=2)

        rate = constant * scan_range ** (slope)

        return numpy.vstack((scan_range, rate)).transpose()

    @property
    def _color_dict(self):
        if not self._color_dict_:
            fix_map = getattr(self._model_map, self.scan_results.fixed)
            relavent_reactions = fix_map.isProductOf() + \
                                 fix_map.isSubstrateOf() + \
                                 fix_map.isModifierOf()
            num_of_cols = len(relavent_reactions) + 3
            cmap = get_cmap('Set2')(
                numpy.linspace(0, 1.0, num_of_cols))[:, :3]
            color_list = [rgb_to_hsv(*cmap[i, :]) for i in range(num_of_cols)]

            relavent_reactions.sort()
            color_dict = dict(
                zip(['Total Supply'] +
                    ['J_' + reaction for reaction in relavent_reactions] +
                    ['Total Demand'],
                    color_list))
            # just to darken the colors a bit
            for k, v in color_dict.iteritems():
                color_dict[k] = [v[0], 1, v[2]]

            self._color_dict_ = color_dict

        return self._color_dict_

    def _make_flux_ld(self):
        color_dict = self._color_dict

        flux_ld_dict = {}

        demand_blocks = ['J_' + dem_reac for dem_reac in getattr(
            self._model_map, self.scan_results.fixed).isSubstrateOf()]
        supply_blocks = ['J_' + sup_reac for sup_reac in getattr(
            self._model_map, self.scan_results.fixed).isProductOf()]

        for flux in self.scan_results.flux_names:
            flux_col = self.scan_results.flux_names.index(flux)
            x_data = self.scan_results.scan_range
            y_data = self.scan_results.flux_data[:, flux_col]
            latex_expr = self._ltxe.expression_to_latex(flux)
            flux_color = self._color_dict[flux]
            color = hsv_to_rgb(flux_color[0],
                               flux_color[1],
                               flux_color[2] * 0.9)
            for dem in demand_blocks:
                if dem == flux:
                    flux_ld_dict[flux] = \
                        LineData(name=flux,
                                 x_data=x_data,
                                 y_data=y_data,
                                 categories=['Fluxes',
                                             'Demand',
                                             flux],
                                 properties={'label': '$%s$' % latex_expr,
                                             'color': color})
                    break
            for sup in supply_blocks:
                if sup == flux:
                    flux_ld_dict[flux] = \
                        LineData(name=flux,
                                 x_data=x_data,
                                 y_data=y_data,
                                 categories=['Fluxes',
                                             'Supply',
                                             flux],
                                 properties={'label': '$%s$' % latex_expr,
                                             'color': color})
                    break
        self._flux_ld_dict = flux_ld_dict

    def _make_ec_ld(self):
        ec_ld_dict = {}

        for ec_name in self.scan_results.ec_names:
            for flux, flux_ld in self._flux_ld_dict.iteritems():
                ec_reaction = flux[2:]
                if 'ec' + ec_reaction + '_' + self.scan_results.fixed in ec_name:
                    flux_color = self._color_dict[flux]
                    color = hsv_to_rgb(flux_color[0],
                                       flux_color[1] * 0.5,
                                       flux_color[2])
                    ec_data = self.scan_results[ec_name]
                    categories = ['Elasticity Coefficients'] + \
                                 flux_ld.categories[1:]
                    latex_expr = self._ltxe.expression_to_latex(ec_name)

                    ec_ld_dict[ec_name] = \
                        LineData(name=ec_name,
                                 x_data=ec_data[:, 0],
                                 y_data=ec_data[:, 1],
                                 categories=categories,
                                 properties={'label': '$%s$' % latex_expr,
                                             'color': color})
        self._ec_ld_dict = ec_ld_dict

    def _make_rc_ld(self):
        rc_ld_dict = {}

        for rc_name in self.scan_results.rc_names:
            for flux, flux_ld in self._flux_ld_dict.iteritems():
                rc_flux = 'J' + flux[2:]
                if 'rc' + rc_flux + '_' in rc_name:
                    flux_color = self._color_dict[flux]
                    color = hsv_to_rgb(flux_color[0],
                                       flux_color[1],
                                       flux_color[2] * 0.7)
                    rc_data = self.scan_results[rc_name]
                    categories = ['Response Coefficients'] + \
                                 flux_ld.categories[1:]
                    latex_expr = self._ltxe.expression_to_latex(rc_name)

                    rc_ld_dict[rc_name] = \
                        LineData(name=rc_name,
                                 x_data=rc_data[:, 0],
                                 y_data=rc_data[:, 1],
                                 categories=categories,
                                 properties={'label': '$%s$' % latex_expr,
                                             'color': color,
                                             'ls': '--'})
        self._rc_ld_dict = rc_ld_dict

    def _make_prc_ld(self):

        def get_prc_route(prc, flux, fixed):
            without_prefix = prc.split('prc')[1]
            without_flux = without_prefix.split(flux)[1][1:]
            route = without_flux.split(fixed)[1][1:]
            return route

        prc_ld_dict = {}

        for prc_name in self.scan_results.prc_names:
            for flux, flux_ld in self._flux_ld_dict.iteritems():
                prc_flux = 'J' + flux[2:]
                if 'prc' + prc_flux + '_' + self.scan_results.fixed in prc_name:
                    route_reaction = get_prc_route(prc_name,
                                                   prc_flux,
                                                   self.scan_results.fixed)
                    flux_color = self._color_dict['J_' + route_reaction]
                    color = hsv_to_rgb(flux_color[0],
                                       flux_color[1] * 0.5,
                                       flux_color[2])
                    prc_data = self.scan_results[prc_name]
                    categories = ['Partial Response Coefficients'] + \
                                 flux_ld.categories[1:]
                    latex_expr = self._ltxe.expression_to_latex(prc_name)

                    prc_ld_dict[prc_name] = \
                        LineData(name=prc_name,
                                 x_data=prc_data[:, 0],
                                 y_data=prc_data[:, 1],
                                 categories=categories,
                                 properties={'label': '$%s$' % latex_expr,
                                             'color': color})
        self._prc_ld_dict = prc_ld_dict

    def _make_total_flux_ld(self):
        total_flux_ld_dict = {}
        col = self._color_dict['Total Supply']
        total_flux_ld_dict['Total Supply'] = \
            LineData(name='Total Supply',
                     x_data=self.scan_results.scan_range,
                     y_data=self.scan_results.total_supply,
                     categories=['Fluxes',
                                 'Supply',
                                 'Total Supply'],
                     properties={'label': '$%s$' % 'Total\,Supply',
                                 'color': hsv_to_rgb(col[0], col[1],
                                                     col[2] * 0.9),
                                 'ls': '--'})

        col = self._color_dict['Total Demand']
        total_flux_ld_dict['Total Demand'] = \
            LineData(name='Total Demand',
                     x_data=self.scan_results.scan_range,
                     y_data=self.scan_results.total_demand,
                     categories=['Fluxes',
                                 'Demand',
                                 'Total Demand'],
                     properties={'label': '$%s$' % 'Total\,Demand',
                                 'color': hsv_to_rgb(col[0], col[1],
                                                     col[2] * 0.9),
                                 'ls': '--'})

        self._total_flux_ld_dict = total_flux_ld_dict

    def plot(self):

        category_classes = OrderedDict([
            ('Supply/Demand', [
                'Supply',
                'Demand']),
            ('Reaction Blocks',
             self.scan_results.flux_names +
             ['Total Supply', 'Total Demand']),
            ('Lines', [
                'Fluxes',
                'Elasticity Coefficients',
                'Response Coefficients',
                'Partial Response Coefficients'])])

        line_data_list = [v for v in self._line_data_dict.itervalues()]

        scan_fig = ScanFig(line_data_list,
                           ax_properties={'xlabel': '[%s]' %
                                                    self.scan_results.fixed.replace(
                                                        '_', ' '),
                                          'ylabel': 'Rate',
                                          'xscale': 'log',
                                          'yscale': 'log',
                                          'xlim': [self.scan_results.scan_min,
                                                   self.scan_results.scan_max],
                                          'ylim': [self.scan_results.flux_min,
                                                   self.scan_results.flux_max * 2
                                                   ]},
                           category_classes=category_classes,
                           base_name=self._analysis_method,
                           working_dir=self._working_dir)

        scan_fig.toggle_category('Supply', True)
        scan_fig.toggle_category('Demand', True)
        scan_fig.toggle_category('Fluxes', True)
        scan_fig.ax.axvline(self.scan_results.fixed_ss, ls=':', color='gray')

        return scan_fig

    def plot_decompose(self):
        from warnings import warn, simplefilter
        simplefilter('always', DeprecationWarning)
        warn('plot_decompose has been renamed to `do_mca_scan, use that '
             'method in the future`', DeprecationWarning, stacklevel=1)
        simplefilter('default', DeprecationWarning)
        return self.do_mca_scan()


    @silence_print
    def do_mca_scan(self):
        ecs = []
        ccs = []
        prc_names = []
        rc_names = []
        rc_pos = []

        reagent_of = [each[2:] for each in self.scan_results.flux_names]
        all_reactions = reagent_of + \
                        getattr(self._model_map,
                                self.scan_results.fixed).isModifierOf()

        arl = len(all_reactions)
        strt = 0
        stp = arl
        for flux_reaction in self.scan_results.flux_names:
            reaction = flux_reaction[2:]
            rc_names.append('rcJ%s_%s' % (reaction, self.scan_results.fixed))

            rc_pos.append(range(strt, stp))
            strt += arl
            stp += arl

            for route_reaction in all_reactions:
                ec = 'ec' + route_reaction + '_' + self.scan_results.fixed
                cc = 'ccJ' + reaction + '_' + route_reaction

                name = 'prcJ%s_%s_%s' % (reaction,
                                         self.scan_results.fixed,
                                         route_reaction)

                # ecs.append(ec)
                if ec not in ecs:
                    ecs.append(ec)
                ccs.append(cc)
                prc_names.append(name)

        ec_len = len(ecs)

        user_output = [self.scan_results.fixed] + ecs + ccs


        scanner = pysces.Scanner(self.mod)
        scanner.quietRun = True
        scanner.addScanParameter(self.scan_results.fixed,
                                 self.scan_results.scan_min,
                                 self.scan_results.scan_max,
                                 self.scan_results.scan_points,
                                 log=True)
        scanner.addUserOutput(*user_output)
        scanner.Run()

        ax_properties = {'ylabel': 'Coefficient Value',
                         'xlabel': '[%s]' %
                                   self.scan_results.fixed.replace('_', ' '),
                         'xscale': 'log',
                         'yscale': 'linear',
                         'xlim': [self.scan_results.scan_min,
                                  self.scan_results.scan_max]}

        cc_ec_data_obj = Data2D(mod=self.mod,
                                column_names=user_output,
                                data_array=scanner.UserOutputResults,
                                ltxe=self._ltxe,
                                analysis_method=self._analysis_method,
                                ax_properties=ax_properties,
                                file_name='cc_ec_scan',
                                num_of_groups=ec_len,
                                working_dir=path.split(self._working_dir)[0])

        rc_data = []

        all_outs = scanner.UserOutputResults[:, 1:]

        ec_outs = all_outs[:, :ec_len]
        cc_outs = all_outs[:, ec_len:]
        ec_positions = range(ec_len) * (len(prc_names)/ec_len)


        for i, prc_name in enumerate(prc_names):

            ec_col_data = ec_outs[:, ec_positions[i]]

            cc_col_data = cc_outs[:, i]
            # ec_col_data = outs[:, i]
            # cc_col_data = outs[:, i + cc_s_pos]
            col = ec_col_data * cc_col_data
            rc_data.append(col)

        temp = numpy.vstack(rc_data).transpose()
        rc_data += [numpy.sum(temp[:, rc_pos[i]], axis=1) for i in
                    range(len(rc_names))]

        rc_out_arr = [scanner.UserOutputResults[:, 0]] + rc_data
        rc_out_arr = numpy.vstack(rc_out_arr).transpose()
        rc_data_obj = Data2D(mod=self.mod,
                             column_names=[self.scan_results.fixed] + prc_names + rc_names,
                             data_array=rc_out_arr,
                             ltxe=self._ltxe,
                             analysis_method=self._analysis_method,
                             ax_properties=ax_properties,
                             file_name='prc_scan',
                             num_of_groups=ec_len,
                             working_dir=path.split(self._working_dir)[0])
        #rc_data_obj._working_dir = path.split(self._working_dir)[0]
        #cc_ec_data_obj._working_dir = path.split(self._working_dir)[0]

        return rc_data_obj, cc_ec_data_obj




import copy
import warnings
from os import path

import numpy as np
import pandas as pd
from pysces import Scanner, ParScanner
from pysces import model as pysces_model
from matplotlib import pyplot as plt

from .misc import scanner_range_setup, DotDict, cc_dict, rc_dict, ec_dict, prc_dict, is_parameter, is_species, \
    is_mca_coef, silence_print

__all__ = ['compare_models',
           'SteadyStateComparer',
           'SimulationComparer',
           'ParameterScanComparer',
           'ClosedOpenComparer']

def compare_models(model_list, model_mapping = None, augment_mapping=True, comparison_type = 'ss'):
    """
    Returns a specified model comparer for a list of models.

    The returned model comparer object considers the first model in
    `model_list` as the "base model" to which the other models are
    compared. Since models may have a different naming scheme for
    reactions, species, etc. `model_mapping` allows the user to specify
    the names of attributes of other models as they relate to the base
    model's attribute names.

    Parameters
    ----------
    model_list : list of PysMod
        A list of PysMod (Pysces model) objects to be compared.
    model_mapping : array, optional (Default : None)
        An array that maps the names attributes of the base model to
        others. Each column relates to a different model in `model_list`
        in order. The first column contains the names of attributes of
        the base model, whereas each other column contains the name as
        it appears in the other models. Only species, reactions and
        parameters are needed as other attributes (e.g. control
        coefficients) are derived from these three categories. If None
        is supplied, all attributes are assumed to be named in the
        same way.
    augment_mapping : boolean, optional (Default : True)
        This parameter specifies if `model_mapping` should be used to
        augment a model_mapping which assumes all attributes are named
        in the same way. This is useful if only a few model attributes
        differ in their naming. If set to false, `model_mapping` must
        specify the relationships between all species, reaction and
        parameter names.
    comparison_type : str, optional (Default : "ss")
        A string that specifies which type of comparison should be made.
        Options are 'ss' for SteadyStateComparer, "parscan" for
        ParameterScanComparer, 'sim' for SimulationComparer, and
        'closed_open' for 'ClosedOpenComparer'.

    Returns
    -------
    values: ModelComparer
        A model comparer object instantiated with the arguments supplied
        above.
    """
    if comparison_type == 'ss':
        comparer = SteadyStateComparer
    elif comparison_type == 'parscan':
        comparer = ParameterScanComparer
    elif comparison_type == 'sim':
        comparer = SimulationComparer
    elif comparison_type == 'closed_open':
        comparer = ClosedOpenComparer
    else:
        raise AssertionError, 'Incorrect comparison_type'
    return  comparer(model_list, model_mapping, augment_mapping)


class ModelMapper(object):
    @staticmethod
    def map_models(model_list, model_mapping=None, augment_mapping=True):
        # Replace duplicate models with clones
        model_list = ModelMapper.replace_with_clones(model_list)
        base_model = model_list[0]

        # Generate unique model names
        model_names = ModelMapper.generate_unique_model_names(model_list)

        if augment_mapping and model_mapping is not None:
            augment_map_dict_list = ModelMapper.make_map_dict_list(model_mapping)
            model_mapping = None
        else:
            augment_map_dict_list = None

        # Generate model mapping if needed
        if model_mapping is None:
            model_mapping = ModelMapper.auto_map_models(model_list)

        # Ensure that supplied (or generated) model mapping
        # at least has the correct shape
        assert model_mapping.shape[1] == len(model_list)

        # Make list of model_dicts
        map_dict_list = ModelMapper.make_map_dict_list(model_mapping)

        if augment_map_dict_list:
            for map_dict, augment_map_dict in zip(map_dict_list,
                                                  augment_map_dict_list):
                map_dict.update(augment_map_dict)

        # Get newly fixed parameters for other models
        fixed_param_list = [None]
        for i, map_dict in enumerate(map_dict_list[1:]):
            i = 1 + i
            fixed_params = ModelMapper.get_now_fixed(base_model,
                                                     model_list[i],
                                                     map_dict)
            fixed_param_list.append(fixed_params)

        # Add mca coefficients to model_dicts
        base_mca_dict = ModelMapper.make_mca_dict(base_model)
        for map_dict in map_dict_list:
            mca_map_dict = ModelMapper.get_mca_dict_mapping(base_mca_dict,
                                                            map_dict)
            map_dict.update(mca_map_dict)

        model_map_params = zip(model_list, model_names,
                               map_dict_list, fixed_param_list)

        model_maps = []
        for model, model_name, map_dict, fixed_params in model_map_params:
            model_maps.append(ModelMap(model, model_name,
                                       map_dict, fixed_params))

        return model_maps

    @staticmethod
    def make_map_dict_list(mapping_array):
        """
        For a model mapping array, returns a list of dictionaries where
        the keys correspond to the fist column and the values correspond to
        each column (including the first). A 4 column array will thus produce
        a list of dicts of len 4 with the first having the same keys
        as values.
        """
        base_vals = mapping_array[:, 0]
        comp_vals = mapping_array[:, 1:]
        dict_mapping = [dict(zip(base_vals, base_vals))]
        for col in comp_vals.T:
            dict_mapping.append(dict(zip(base_vals, col)))
        return dict_mapping

    @staticmethod
    def replace_with_clones(model_list):
        """
        For a list of models potentially containing duplicate PySMod
        instantiations, returns a list of unique objects.
        """
        new_model_list = []
        for model in model_list:
            if model in new_model_list:
                new_model_list.append(ModelMapper.clone_model(model))
            else:
                new_model_list.append(model)
        return new_model_list

    @staticmethod
    def generate_unique_model_names(model_list):
        """
        Returns a list of unique model names for a list of models.
        """
        model_names = [path.split(model.ModelFile)[-1][:-4] \
                       for model in model_list]

        for i, model_name in enumerate(model_names):
            indices = [j for j, v in enumerate(model_names) \
                       if v == model_name]
            if len(indices) > 1:
                for j, index in enumerate(indices):
                    model_names[index] = model_name + '_' + str(j)
        return model_names

    @staticmethod
    def auto_map_models(model_list):
        """
        For a list of models returns an array where each column maps contains
        the names of model attributes as they appear in each model. Missing
        species, parameters, and reactions are replaced with None.
        """
        base_model = model_list[0]
        comparison_models = model_list[1:]

        base_attributes = list(base_model.species) + \
                          list(base_model.reactions) + \
                          list(base_model.parameters)

        mapping_array = [base_attributes]
        for model in comparison_models:
            match_attributes = [attr if hasattr(model, attr) \
                                    else None \
                                for attr in base_attributes]
            mapping_array.append(match_attributes)
        return np.array(mapping_array).T

    @staticmethod
    def clone_model(model):
        """
        Given a model this method returns a new instantiation of the same
        model.

        See also:
        replace_with_clones
        """
        new_model_path = path.join(model.ModelDir,
                                   model.ModelFile)
        new_model = pysces_model(new_model_path)
        return new_model

    @staticmethod
    def get_mca_dict_mapping(mca_dict, map_dict):
        """
        Given an mca dictionary of the base model and a map dictionary of any
        comparison model, this returns a dictionary that maps the names of
        mca coefficients in the base model to those in the comparison.

        See also:
        get_equiv_coeff
        """
        return {k: ModelMapper.get_equiv_coeff(k, mca_dict, map_dict)
                for k in mca_dict.iterkeys()}

    @staticmethod
    def get_equiv_coeff(coeff, mca_dict, map_dict):
        """
        Given an mca coefficient, an mca dictionary of the base model and a map
        dictionary of any comparison model, this returns the name of the
        equivalent mca coefficient in the comparison model.

        See also:
        get_mca_dict_mapping
        """
        coeff_type = ModelMapper.get_coeff_type(coeff)
        coeff_components = mca_dict[coeff]
        equiv_components = [map_dict[component] for component \
                            in coeff_components]
        if None in equiv_components:
            return None
        else:
            if coeff_type in ['cc', 'prc', 'rc']:
                prefix = coeff_type + 'J'
            else:
                prefix = coeff_type
            if coeff_type == 'prc':
                return prefix + '{}_{}_{}'.format(*equiv_components)
            else:
                return prefix + '{}_{}'.format(*equiv_components)

    @staticmethod
    def make_mca_dict(model):
        """
        Returns a dictionary with the mca coefficients of a model as keys
        and a tuple of the components that make up each coefficient as values.
        """
        full_dict = {}
        full_dict.update(ec_dict(model))
        full_dict.update(cc_dict(model))
        full_dict.update(rc_dict(model))
        full_dict.update(prc_dict(model))
        return full_dict

    @staticmethod
    def get_coeff_type(coefficient):
        """
        Returns the coefficient type.
        """
        if coefficient.startswith('prc'):
            return 'prc'
        elif coefficient.startswith('rc'):
            return 'rc'
        elif coefficient.startswith('ec'):
            return 'ec'
        elif coefficient.startswith('cc'):
            return 'cc'

    @staticmethod
    def get_now_fixed(base_model, comparison_model, map_dict):
        now_fixed_list = []
        for k, v in map_dict.iteritems():
            if is_species(k, base_model) and is_parameter(v, comparison_model):
                now_fixed_list.append(k)
        return now_fixed_list


class ModelMap(object):
    def __init__(self, model, model_name, attr_dict, is_now_fixed=None):
        self.model = model
        self.model_name = model_name
        self.attr_dict = attr_dict
        self._rev_attr_dict = ModelMap._reverse_dict(self.attr_dict)
        if not is_now_fixed:
            is_now_fixed = []
        self.is_now_fixed = is_now_fixed

    @staticmethod
    def _reverse_dict(dict_):
        return {v: k for k, v in dict_.iteritems()}

    def getattrname(self, key):
        prefix, suffix, key = ModelMap._prefix_suffix_getter(key)
        attr = self.attr_dict[key]
        if attr is None:
            return None
        elif key in self.is_now_fixed:
            return attr
        else:
            return prefix + attr + suffix

    def hasattr(self, item):
        _, _, item = ModelMap._prefix_suffix_getter(item)
        if item in self.attr_dict.keys():
            return True
        else:
            return False

    def setattr(self, key, value):
        if self.hasattr(key):
            true_key = self.getattrname(key)
            if true_key is not None and hasattr(self.model, true_key):
                setattr(self.model, true_key, value)

    def getattr(self, item):
        true_key = self.getattrname(item)
        if true_key is None:
            return np.NaN
        else:
            return getattr(self.model, true_key)

    def attr_names_from_base_names(self, key_list, include_nones=False):
        converted_attrs = []
        for key in key_list:
            true_attr = self.getattrname(key)
            if true_attr is None:
                if include_nones:
                    converted_attrs.append(true_attr)
            else:
                converted_attrs.append(true_attr)
        return converted_attrs

    def base_names_from_attr_names(self, attr_list, add_ss = False):
        new_attr_list = [ModelMap._prefix_suffix_getter(attr) for \
                         attr in attr_list]
        for i, p_s_a in enumerate(new_attr_list):
            prefix, suffix, attr = p_s_a
            base_attr = self._rev_attr_dict[attr]
            if base_attr in self.is_now_fixed and add_ss:
                suffix = '_ss'
                new_attr_list[i] = (prefix, suffix, attr)

        converted_attrs = [prefix + self._rev_attr_dict[attr] + suffix for \
                           prefix, suffix, attr in new_attr_list]
        return converted_attrs

    def setattrs_from_dict(self, attr_dict):
        for key, value in attr_dict.iteritems():
            self.setattr(key, value)

    def getattrs_from_list(self, attr_list):
        return [self.getattr(key) for key in attr_list]

    @staticmethod
    def _prefix_suffix_getter(key):
        prefix = ''
        suffix = ''
        flux_prefix = 'J_'
        ssconc_suffix = '_ss'
        init_prefix = '_init'
        if key.startswith(flux_prefix):
            key = key[2:]
            prefix = flux_prefix
        elif key.endswith(ssconc_suffix):
            key = key[:-3]
            suffix = ssconc_suffix
        elif key.endswith(init_prefix):
            key = key[:-5]
            suffix = init_prefix
        return prefix, suffix, key


class ResultsGenerator(object):
    @staticmethod
    def percentage_change(a, b, abs_values = False):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if abs_values:
                return np.divide(np.abs(np.subtract(b, a)), np.abs(a)) * 100
            else:
                return np.divide(np.subtract(b, a), a) * 100

    @staticmethod
    def do_par_scan(model, scan_in, scan_out, start, end, points, is_log_range, par_scan, par_engine):
        if par_scan:
            scanner = ParScanner(model, engine=par_engine)
        else:
            scanner = Scanner(model)
        scan_out = [str(each) for each in scan_out]
        scanner.addScanParameter(scan_in, start=start, end=end, points=points, log=is_log_range)
        scanner.addUserOutput(*scan_out)
        scanner.Run()
        result = scanner.getResultMatrix()
        return result

    @staticmethod
    def do_simulation(model, time_range, sim_out):
        model.sim_time = time_range
        model.Simulate(userinit=1)
        results, labels = model.data_sim.getAllSimData(lbls=True)
        sim_out = ['Time'] + sim_out
        sim_out_index = [labels.index(out_label) for out_label in sim_out]
        results_to_keep = results[:,sim_out_index]
        return results_to_keep


class BaseModelComparer(object):
    def __init__(self, model_list, model_mapping=None, augment_mapping=True):
        super(BaseModelComparer, self).__init__()
        assert len(model_list) > 1, 'Two or more models are needed to make a comparison'

        self.mmap_list = ModelMapper.map_models(model_list=model_list,
                                                model_mapping=model_mapping,
                                                augment_mapping=augment_mapping)

        self.raw_data = None
        self.comparison = None

    @silence_print
    def do_compare(self, output_list=None, custom_init=None, uniform_init=True):
        """
        Performs the comparison of the models.

        Results are stored in `self.raw_data` and `self.comparison`.

        Parameters
        ----------
        output_list : list of str, optional (Default : None)
            A list of model attributes to compare. Valid options are:
            species, reactions, parameters, and mca coefficients.
        custom_init : dict or list of dict, optional (Default : None)
            A dictionary specifying the initial conditions of the models.
            Keys specify model attributes (in str) and values are float
            values of these attributes. A list of dictionaries can also
            be supplied in order to give each model a different set of
            initial conditions. The list follows the same order as
            the `model_list` supplied during object instantiation.
        uniform_init : boolean, optional (Default : True)
            If set to True, the attributes of each model will be set to
            that of the base model prior to comparison. `custom_inits`
            are applied after making model attributes uniform.
        """
        if output_list is None:
            output_list = self._get_default_outputs()

        self._uniform_init(uniform_init)
        self._custom_init(custom_init)
        self._generate_raw_data(output_list)
        self._compare_raw_data()

    @silence_print
    def _get_default_outputs(self):
        base_model = self.mmap_list[0].model
        return list(base_model.species) + list(base_model.reactions)

    @silence_print
    def _custom_init(self, custom_init):
        if custom_init:
            if type(custom_init) is dict:
                for mmap in self.mmap_list:
                    mmap.setattrs_from_dict(custom_init)
            elif type(custom_init) is list:
                num_of_models = len(self.mmap_list)
                all_dicts = all([type(each) is dict for each in custom_init])
                correct_num_of_dicts = len(custom_init) == num_of_models
                assert all_dicts and correct_num_of_dicts, \
                    "list of custom inits must contain {} dictionaries".format(num_of_models)
                for custom_init_dict, mmap in zip(custom_init, self.mmap_list):
                    mmap.setattrs_from_dict(custom_init_dict)
            else:
                raise AssertionError, "custom_inits must be a dictionary or a list of dictionaries"

    @silence_print
    def _uniform_init(self, uniform_init):
        if uniform_init:
            base_model = self.mmap_list[0].model
            base_parameter_dict = {k:getattr(base_model, k) for k in base_model.parameters}
            base_species_dict = {k:getattr(base_model, k) for k in base_model.species}

            base_attribute_dict = {}
            base_attribute_dict.update(base_parameter_dict)
            base_attribute_dict.update(base_species_dict)

            for mmap in self.mmap_list[1:]:
                mmap.setattrs_from_dict(base_attribute_dict)

    def _compare_raw_data(self):
        raise NotImplementedError

    def _generate_raw_data(self, output_list):
        raise NotImplementedError

    @silence_print
    def _output_to_ss(self, output_list):
        base_model = self.mmap_list[0].model
        new_out_list = []
        for each in output_list:
            if each in base_model.species:
                new_out_list.append(each + '_ss')
            elif each in base_model.reactions:
                new_out_list.append('J_' + each)
            else:
                new_out_list.append(each)
        return new_out_list


class SteadyStateComparer(BaseModelComparer):
    @silence_print
    def _generate_raw_data(self, output_list):
        output_list = self._output_to_ss(output_list)
        all_results = []

        base_model = self.mmap_list[0].model
        state_method = 'mca' if any([is_mca_coef(attr, base_model) for attr in output_list]) else 'ss'

        for mmap in self.mmap_list:
            if state_method == 'mca':
                mmap.model.doMca()
            else:
                mmap.model.doState()
            model_results = [mmap.getattr(attr) for attr in output_list]
            all_results.append(model_results)
        all_results = np.array(all_results).T
        self.raw_data = pd.DataFrame(data=all_results,
                                     columns=[mmap.model_name for mmap in self.mmap_list],
                                     index=output_list)

    @silence_print
    def _compare_raw_data(self):
        raw_data = self.raw_data
        data_matrix = raw_data.as_matrix()
        reference_values = data_matrix[:,0]
        change_values = data_matrix[:,1:]

        comparison_names = []
        all_results = []
        for i, change_col in enumerate(change_values.T):
            comparison_name = '{}_{}'.format(self.mmap_list[0].model_name,
                                             self.mmap_list[i+1].model_name)

            comparison_names.append(comparison_name)
            comparison_result = ResultsGenerator.percentage_change(reference_values,
                                                                   change_col)
            all_results.append(comparison_result)

        all_results = np.array(all_results).T
        self.comparison = pd.DataFrame(data=all_results,
                                       columns=comparison_names,
                                       index=raw_data.index)


class ParameterScanComparer(BaseModelComparer):
    @silence_print
    def do_compare(self, scan_in, scan_range, output_list=None, custom_init=None,
                   uniform_init=True, par_scan=False, par_engine='multiproc'):
        """
        Performs the comparison of the models.

        Results are stored in `self.raw_data` and `self.comparison`.

        Parameters
        ----------
        scan_in : str
            A string that specifies the parameter to be varied
        scan_range : array-like
            An array that specifies the range of values over which
            `scan_in` should be varied
        output_list : list of str, optional (Default : None)
            A list of model attributes to compare. Valid options are:
            species, reactions, parameters, and mca coefficients.
        custom_init : dict or list of dict, optional (Default : None)
            A dictionary specifying the initial conditions of the models.
            Keys specify model attributes (in str) and values are float
            values of these attributes. A list of dictionaries can also
            be supplied in order to give each model a different set of
            initial conditions. The list follows the same order as
            the `model_list` supplied during object instantiation.
        uniform_init : boolean, optional (Default : True)
            If set to True, the attributes of each model will be set to
            that of the base model prior to comparison. `custom_inits`
            are applied after making model attributes uniform.
        par_scan : boolean, optional (Default : False)
            If True the parameter scan will be executed using parallel
            processing.
        par_enging: str, optional (Default : 'multiproc')
            The parallel engine to be used. Options are dictated by
            PySCeS's ParScanner.
        """
        if output_list is None:
            output_list = self._get_default_outputs()
        self._uniform_init(uniform_init)
        self._custom_init(custom_init)

        self._generate_raw_data(scan_in, output_list, scan_range, par_scan, par_engine)
        self._compare_raw_data()

    @silence_print
    def _generate_raw_data(self, scan_in, output_list, scan_range, par_scan, par_engine):
        output_list = self._output_to_ss(output_list)
        main_column_labels = [scan_in] + output_list
        start, end, points, is_log_range = scanner_range_setup(scan_range)

        all_results = DotDict()
        for i, mmap in enumerate(self.mmap_list):
            current_scan_in = mmap.getattrname(scan_in)
            current_scan_out = mmap.attr_names_from_base_names(output_list)
            current_col_labels = [scan_in] + mmap.base_names_from_attr_names(current_scan_out, add_ss=True)

            # parameter scan
            raw_result = ResultsGenerator.do_par_scan(model=mmap.model,
                                                      scan_in=current_scan_in,
                                                      scan_out=current_scan_out,
                                                      start=start,
                                                      end=end,
                                                      points=points,
                                                      is_log_range=is_log_range,
                                                      par_scan=par_scan,
                                                      par_engine=par_engine)

            complete_results = pd.DataFrame(data=raw_result,
                                            columns=current_col_labels)
            complete_results = complete_results.reindex(columns=main_column_labels)
            complete_results = complete_results.set_index(scan_in)
            all_results[mmap.model_name] = complete_results
        self.raw_data = all_results

    @silence_print
    def _compare_raw_data(self):
        raw_data = self.raw_data
        reference_model_name = self.mmap_list[0].model_name
        reference_matrix = raw_data[reference_model_name].reset_index().as_matrix()
        column_labels = raw_data[reference_model_name].reset_index().columns

        all_results = DotDict()
        for mmap in self.mmap_list[1:]:
            current_model_name = mmap.model_name
            comparison_name = '{}_{}'.format(reference_model_name,
                                             current_model_name)
            change_matrix = raw_data[current_model_name].reset_index().as_matrix()

            comparison_result = ResultsGenerator.percentage_change(reference_matrix[:,1:],change_matrix[:,1:])
            comparison_result = np.hstack([reference_matrix[:,0][:,np.newaxis],comparison_result])
            df = pd.DataFrame(data=comparison_result,
                              columns=column_labels)
            index = df.columns[0]
            df = df.set_index(index)
            all_results[comparison_name] = df
        self.comparison = all_results


    def plot(self,cols=3, width=None, height=None):
        """
        Produces a simple plot that allows for the visual comparison of
        results for each model.


        Parameters
        ----------
        cols : int, optional (Default : 3)
            The number of columns to use when plotting subplots.
        width : float, optional (Default : 15.0)
            The width of the figure in inches.
        height : float, optional (Default : 5.0*number of subplot rows)
            The height of the figure in inches. By default the height is
            dynamic in will be adjusted based on the number of rows of
            subplots where each row is assigned 5 inches.
        Returns
        -------
        f: matplotlib figure
            A matplotlib figure on which results are plotted.
        axxarr: 2-dimensional numpy array of AxesSubplots
            An array that contains the axes associate with the figure.
        """
        columns = self.raw_data.values()[0].columns
        index = self.raw_data.values()[0].index

        assert cols >= 2
        rows = int(np.ceil(len(columns)/float(cols)))

        if not height:
            height = 5.*rows
        if not width:
            width = 15.

        f, axarr = plt.subplots(rows,cols)
        f.set_size_inches(w=width, h=height)

        for i in xrange(cols*rows - len(columns)):
            f.delaxes(axarr.flat[-1-i])

        for column, ax in zip(columns,axarr.flat):
            for k, v in self.raw_data.iteritems():
                ax.plot(index, v[column], '-',label=k)
            ax.set_xlabel(index.name)
            ax.set_title(column)
            ax.legend()
        return f,axarr


class SimulationComparer(ParameterScanComparer):
    @silence_print
    def do_compare(self, time_range, output_list=None, custom_init=None,
                   uniform_init=True):
        """
        Performs the comparison of the models.

        Results are stored in `self.raw_data` and `self.comparison`.

        Parameters
        ----------
        time_range : array-like
            An array that specifies the range of values over which
            `Time` should be varied. It allows for starting points other
            than zero.
        output_list : list of str, optional (Default : None)
            A list of model attributes to compare. Valid options are:
            species, reactions, parameters, and mca coefficients.
        custom_init : dict or list of dict, optional (Default : None)
            A dictionary specifying the initial conditions of the models.
            Keys specify model attributes (in str) and values are float
            values of these attributes. A list of dictionaries can also
            be supplied in order to give each model a different set of
            initial conditions. The list follows the same order as
            the `model_list` supplied during object instantiation.
        uniform_init : boolean, optional (Default : True)
            If set to True, the attributes of each model will be set to
            that of the base model prior to comparison. `custom_inits`
            are applied after making model attributes uniform.
        """
        if output_list is None:
            output_list = self._get_default_outputs()
        self._uniform_init(uniform_init)
        self._custom_init(custom_init)

        self._generate_raw_data(output_list, time_range)
        self._compare_raw_data()

    @silence_print
    def _generate_raw_data(self, output_list, time_range):
        main_column_labels = ['Time'] + output_list
        all_results = DotDict()
        for i, mmap in enumerate(self.mmap_list):

            now_fixed_indicies = []
            for i,out in enumerate(output_list):
                if out in mmap.is_now_fixed:
                    now_fixed_indicies.append(i)
            new_output_list = copy.deepcopy(output_list)
            param_cols = []
            for ind in reversed(now_fixed_indicies):
                out_val = mmap.getattr(output_list[ind])
                new_output_list.pop(ind)
                param_cols.append([out_val for _ in xrange(time_range.shape[0])])
            param_cols = param_cols

            current_sim_out = mmap.attr_names_from_base_names(new_output_list)

            real_sim_out = mmap.attr_names_from_base_names(output_list)
            current_col_labels = ['Time'] + mmap.base_names_from_attr_names(real_sim_out )

            # parameter scan
            raw_result = ResultsGenerator.do_simulation(model=mmap.model,
                                                        time_range=time_range,
                                                        sim_out=current_sim_out)

            new_raw_result = [col for col in raw_result.T]
            for ind, col in zip(reversed(now_fixed_indicies), param_cols):
                new_raw_result.insert(ind, col)
            new_raw_result = np.array(new_raw_result).T

            complete_results = pd.DataFrame(data=new_raw_result,
                                            columns=current_col_labels)

            complete_results = complete_results.reindex(columns=main_column_labels)
            complete_results = complete_results.set_index('Time')
            all_results[mmap.model_name] = complete_results
        self.raw_data = all_results


class ClosedOpenComparer(SimulationComparer):
    @silence_print
    def do_compare(self, time_range, output_list=None, custom_init=None,
                   uniform_init=True):
        """
        Performs the comparison of the models.

        Results are stored in `self.raw_data` and `self.comparison`.

        Parameters
        ----------
        time_range : array-like
            An array that specifies the range of values over which
            `Time` should be varied. It allows for starting points other
            than zero.
        output_list : list of str, optional (Default : None)
            A list of model attributes to compare. Valid options are:
            species, reactions, parameters, and mca coefficients.
        custom_init : dict or list of dict, optional (Default : None)
            A dictionary specifying the initial conditions of the models.
            Keys specify model attributes (in str) and values are float
            values of these attributes. A list of dictionaries can also
            be supplied in order to give each model a different set of
            initial conditions. The list follows the same order as
            the `model_list` supplied during object instantiation.
        uniform_init : boolean, optional (Default : True)
            If set to True, the attributes of each model will be set to
            that of the base model prior to comparison. `custom_inits`
            are applied after making model attributes uniform.
        """
        if output_list is None:
            output_list = self._get_default_outputs()
        self._uniform_init(uniform_init)
        self._custom_init(custom_init)

        self._generate_raw_data(output_list, time_range)
        self._compare_raw_data()

    @silence_print
    def _generate_raw_data(self, output_list, time_range):
        base_mmap = self.mmap_list[0]
        full_scan_column_names = []
        for mmap in self.mmap_list:
            full_scan_column_names = full_scan_column_names + mmap.is_now_fixed
        full_scan_column_names = list(set(full_scan_column_names))
        base_sim_out_list = full_scan_column_names + [out for out in output_list \
                                                      if out not in full_scan_column_names]
        base_column_names = ['Time'] + base_sim_out_list

        all_results = DotDict()
        simulation_results = ResultsGenerator.do_simulation(model=base_mmap.model,
                                                            time_range=time_range,
                                                            sim_out=base_sim_out_list)
        base_complete_results = pd.DataFrame(data=simulation_results,
                                             columns=base_column_names)
        base_complete_results = base_complete_results.set_index('Time')
        all_results[base_mmap.model_name] = base_complete_results

        for i, mmap in enumerate(self.mmap_list[1:]):
            scan_column_names = mmap.is_now_fixed
            output_partial = [out for out in output_list if out not in scan_column_names]
            outputs = self._output_to_ss(output_partial)
            output_col_names = ['Time'] + scan_column_names + output_partial

            time_col = base_complete_results.reset_index()['Time'].as_matrix()
            input_val_list = [base_complete_results[col_name].as_matrix() for col_name in scan_column_names]
            raw_results = []
            for input_vals in zip(*input_val_list):
                current_row = [] + list(input_vals)
                dict_to_set = dict(zip(scan_column_names, input_vals))
                mmap.setattrs_from_dict(dict_to_set)
                mmap.model.doState()
                current_row = current_row + mmap.getattrs_from_list(outputs)
                raw_results.append(current_row)
            raw_results = np.hstack((time_col[:,np.newaxis],np.array(raw_results)))
            complete_results = pd.DataFrame(data=raw_results,
                                            columns=output_col_names)
            complete_results = complete_results.reindex(columns=base_column_names)
            complete_results = complete_results.set_index('Time')
            all_results[mmap.model_name] = complete_results
        self.raw_data = all_results





from setuptools import setup, find_packages
from os import path
try:
  from jupyterpip import cmdclass
except:
  import pip
  import importlib
  pip.main(['install', 'jupyter-pip'])
  cmdclass = importlib.import_module('jupyterpip').cmdclass

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()

packages = find_packages()

setup(
    name='PyscesToolbox',
    version='0.8.9post1',
    packages=packages,
    url='https://github.com/PySCeS/PyscesToolbox',
    download_url='http://github.com/PySCeS/PyscesToolbox/archive/0.8.9.tar.gz',
    license='BSD-3-Clause',
    author='Carl Christensen',
    author_email='carldc@sun.ac.za',
    description='A set of metabolic model analysis tools for PySCeS.',
    long_description=long_description,
    install_requires=['IPython>=4.0.0',
                      'numpy',
                      'sympy',
                      'pysces',
                      'matplotlib',
                      'numpydoc',
                      'networkx==1.11',
                      'ipywidgets==6.0.0',
                      'widgetsnbextension==2.0.0',
                      'jupyter-pip',
                      'pandas'
                      ],
    package_data={'d3networkx_psctb': ['widget.js'],},
    keywords=['metabolism','metabolic control analysis','modelling'],
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Intended Audience :: Science/Research',
                 'Topic :: Scientific/Engineering :: Bio-Informatics',
                 'License :: OSI Approved :: BSD License',
                 'Programming Language :: Python :: 2.7'],
    cmdclass=cmdclass('d3networkx_psctb'),
)


from ConfigParser import ConfigParser
from sys import platform, stdout
from os import path, sep, listdir, access, X_OK, environ, pathsep
# from pkg_resources import resource_stream, resource_filename
from pysces import output_dir
from warnings import warn

_DEFAULT_CONFIG = {'Settings': {
    'maxima_path': 'C:\\Program Files?\\Maxima?\\bin\\maxima.bat'}}

_DEFAULT_CONF_NAME = 'default_config.ini'
_USER_CONF_PATH = path.join(output_dir, 'psctb_config.ini')


class ConfigReader:
    _config = None

    @classmethod
    def reload_config(cls):
        cls._config = None

    @classmethod
    def get_config(cls):
        if not cls._config:
            cls._setup_config()
        return cls._config

    @classmethod
    def _setup_config(cls):
        default_conf = ConfigParser()
        default_conf._sections = _DEFAULT_CONFIG
        try:
            if not path.exists(_USER_CONF_PATH):
                ConfigWriter.write_config(_DEFAULT_CONFIG, _USER_CONF_PATH)
                user_conf = ConfigParser()
                user_conf._sections = _DEFAULT_CONFIG
            else:
                user_conf = cls._read_config(_USER_CONF_PATH)
                ConfigChecker.check_config('user configuration',
                                           _USER_CONF_PATH,
                                           user_conf._sections)
        except MissingSection as e:
            solution = ('falling back to default configuration.'
                        '\nDelete configuration file to restore defaults')
            ConfigChecker.warn_user(e, solution)
            user_conf = ConfigParser()
            user_conf.add_section('Settings')
        except MissingSetting as e:
            solution = ('falling back to default configuration.'
                        '\nSpecify setting to avoid using defaults')
            ConfigChecker.warn_user(e, solution)

        cls._config = cls._compose_config(default_conf, user_conf)

        try:
            if platform == 'win32':
                maxima_path_list = PathFinder.find_path_to(
                    cls._config['maxima_path'])
                if not len(maxima_path_list) > 0:
                    raise IOError(2,
                                  'No valid path to specified file',
                                  cls._config['maxima_path'])

                maxima_path = sorted(maxima_path_list)[-1]
                cls._config['maxima_path'] = maxima_path
            else:
                maxima_path = PathFinder.which('maxima')
                if not maxima_path:
                    raise IOError(2,
                                  'Maxima not installed',
                                  'command not found')
                cls._config['maxima_path'] = 'maxima'
        except IOError as e:
            solution = ('Please check that configuration file specifies '
                        'the correct path for Maxima and '
                        'that Maxima is installed correctly before '
                        'attempting to generate new results with SymCA '
                        '(see documentation for details).')
            ConfigChecker.warn_user(e, solution)
            cls._config['maxima_path'] = None

        cls._config['platform'] = platform
        cls._config['stdout'] = stdout

    @staticmethod
    def _read_config(file_or_path):
        conf = ConfigParser()
        if type(file_or_path) is str:
            conf.read(file_or_path)
        else:
            conf.readfp(file_or_path)
        return conf

    @staticmethod
    def _compose_config(default_conf, user_conf):
        conf_dict = {}
        conf_dict.update(default_conf._sections['Settings'])
        conf_dict.update(user_conf._sections['Settings'])
        if '__name__' in conf_dict:
            conf_dict.pop('__name__')
        return conf_dict


class ConfigChecker:
    @staticmethod
    def _has_all_sections(config_name, config_path, config_dict):
        error_string = ('The {config_name} located at\n{config_path}\n'
                        'does not contain the required section\n'
                        '"{section}".')
        for section in _DEFAULT_CONFIG.keys():
            if section not in config_dict:
                raise MissingSection(
                    error_string.format(**locals()))

    @staticmethod
    def _has_all_settings(config_name, config_path, config_dict):
        error_string = ('The {config_name} located at\n{config_path}\n'
                        'does not contain the required setting\n'
                        '"{setting}"\nunder the section\n"{section}".')
        for section in _DEFAULT_CONFIG.keys():
            for setting in _DEFAULT_CONFIG[section].keys():
                if setting not in config_dict[section]:
                    raise MissingSetting(
                        error_string.format(**locals()))

    @staticmethod
    def check_config(config_name, config_path, config_dict):
        ConfigChecker._has_all_sections(config_name, config_path, config_dict)
        ConfigChecker._has_all_settings(config_name, config_path, config_dict)

    @staticmethod
    def _get_exception_name(exception):
        class_name = str(exception.__class__)
        return class_name[:-2].split('.')[-1]

    @staticmethod
    def warn_user(exception, solution):
        warning_string = ('\n\n'
                          'The following error was encountered:\n"{message}"'
                          '\n\n{solution}')
        if type(exception) is IOError:
            exception.message = exception.strerror + ':\n' + exception.filename
        message = ConfigChecker._get_exception_name(exception) + \
            ' - ' + exception.message
        warn(warning_string.format(**locals()))


class ConfigWriter:
    @staticmethod
    def write_config(config_dict, config_path):
        conf = ConfigParser()
        for section, settings in config_dict.iteritems():
            conf.add_section(section)
            for setting, value in settings.iteritems():
                conf.set(section, setting, value)
        with open(config_path, 'w') as f:
            conf.write(f)


class MissingSetting(Exception):
    pass


class MissingSection(Exception):
    pass


class PathFinder:
    @staticmethod
    def find_path_to(wildcard_path):
        path_parts = wildcard_path.split(sep)
        if platform == 'win32':
            new_paths = [path_parts.pop(0) + sep]
        else:
            new_paths = [sep]

        for i, path_part in enumerate(path_parts):
            if '?' in path_part:
                path_part = path_part[:-1]
                new_new_paths = []
                for each_base in new_paths:
                    possible_matches = PathFinder.find_match(each_base,
                                                             path_part)
                    if len(possible_matches) > 0:
                        for match in possible_matches:
                            new_new_paths.append(path.join(each_base, match))
                new_paths = new_new_paths
            else:
                new_new_paths = []
                for each_path in new_paths:
                    new_path = path.join(each_path, path_part)
                    if path.exists(new_path):
                        new_new_paths.append(new_path)
                new_paths = new_new_paths
        return new_paths

    @staticmethod
    def find_match(base_dir, to_match):
        matches = []
        try:
            dirs_in_basedir = listdir(base_dir)
        except OSError:
            dirs_in_basedir = []
        for directory in dirs_in_basedir:
            if to_match in directory:
                matches.append(directory)
        return matches

    @staticmethod
    def which(program):
        def is_exe(fpath):
            return path.isfile(fpath) and access(fpath, X_OK)

        fpath, fname = path.split(program)
        if fpath:
            if is_exe(program):
                return program
        else:
            for epath in environ["PATH"].split(pathsep):
                epath = epath.strip('"')
                exe_file = path.join(epath, program)
                if is_exe(exe_file):
                    return exe_file

        return None



from os import path, mkdir, listdir

from pysces import output_dir


__all__ = ['get_model_name', 'make_path', 'next_suffix', 'get_file_path',
           'get_fmt']


def get_model_name(mod):
    """Returns the file name of a pysces model object sans the file extension.

    Parameters
    ----------
    mod : PysMod
        Model of interest.

    Returns
    -------
    str
        File name of a `mod` sans extension.

    """
    return path.split(mod.ModelFile)[1][:-4]


def next_suffix(directory, base_name, ext=None):
    """Returns the number of the next suffix to be appended to a base file name
    when saving a file.

    This function checks a ``directory`` for files containing ``base_name`` and
    returns a number that is equal to the suffix of a file named ``base_name``
    with the largest suffix plus one.

    Parameters
    ----------
    directory : str
        The directory to inspect for files.
    base_name : str
        The base name (sans suffix) to check for.

    Returns
    -------
    int
        The next suffix to write
    """

    # find all files in dir, exclude subdirs
    files = [each for each in listdir(directory) if
             path.isfile(path.join(directory, each))]
    # start counting at zero
    next_num = 0
    if not ext:
        ext = ''
    for each in files:
        if each.startswith(base_name) and each.endswith(ext):
            start = len(base_name + '_')
            end = start + 1
            num = int(each[start:end])
            if num >= next_num:
                next_num = num + 1
    return next_num


def get_file_path(working_dir, internal_filename, fmt, fixed=None,
                  file_name=None, write_suffix=True ):
    """An heuristic for determining the correct file name.

    This function determines the file name according to the information
    supplied by the user and the internals of a specific class.

    Parameters
    ----------
    working_dir : str
        The working dir of the specific class (where files are saved if no
        file name is supplied)
    internal_filename : str
        The default base name (sans numbered suffix) of files when no other
        details are provided.
    fmt : str
        The format (extension) that the file should be saved in. This is
        used both in determining file name if no file name is provided as
        well as when a file name without extension is provided.
    fixed : str, Optional (Default : None)
        In the case that a metabolite is fixed, files will be saved in a
        subdirectory of the working directory that corresponds to the
        fixed metabolite.
     file_name : str, Optional (Default : None)
        If a file name is supplied it overwrites all other options except
        ``fmt`` in the case where no extension is supplied.
    Returns
    -------
    str
        The final file name

    """
    if not file_name:
        if fixed:
            save_path = path.join(working_dir, fixed)
        else:
            save_path = working_dir
        if not path.exists(save_path):
            mkdir(save_path)
        suffix = ''
        if write_suffix:
            suffix = '_' + str(next_suffix(save_path,
                                     internal_filename,
                                     fmt))
        fname = internal_filename +  suffix + '.' + fmt
        file_name = path.join(save_path, fname)
    else:
        if path.splitext(file_name)[1] == '':
            file_name = file_name + '.' + fmt
        save_path = path.split(file_name)[0]
        if not path.exists(save_path):
            pass
            mkdir(save_path)
    return file_name


def get_fmt(file_name):
    """Gets the extension (fmt) from a file name.

    Parameters
    ----------
    file_name : str
        The file to get an extension from

    Returns
    -------
    str
        The extension string
    """
    return path.splitext(file_name)[1][1:]


def make_path(mod, analysis_method, subdirs=[]):
    """Creates paths based on model name and analysis type.


    This function is used to create directories (in the case where they don't
    already exist) to write analysis  results to and return the path name.
    Subdirectories can also be created.



    /path/to/Pysces/model_name/analysis_method/subdir1/subdir2/

    Parameters
    ----------
    mod : PysMod
        The model being analysed.
    analysis_method : str
        The name of the tool being used to analyse the model.
    subdirs : list of str
        An optional list of subdirectories where each additional entry in the
        list will create a subdirectory in the previous directories.

    Returns
    -------
    str
        The directory string

    Examples
    --------
    >>> print make_path(mod, 'analysis_method', subdirs = ['subdir1', subdir2])
    '/path/to/Pysces/model_name/analysis_method/subdir1/subdir2/'


    """

    model_name = get_model_name(mod)

    model_path = path.join(output_dir,
                           model_name)

    analysis_path = path.join(output_dir,
                              model_name,
                              analysis_method)

    dirs = [model_path, analysis_path]

    for i, each in enumerate(subdirs):
        tpth = path.join(dirs[1 + i], each)
        dirs.append(tpth)

    for each in dirs:
        if not path.exists(each):
            mkdir(each)

    return dirs[-1]
